''' A panel that allows user to choose an exposure preset or to add a new preset
'''

import json
from functools import partial
from loguru import logger

from kivy.app import App
from kivy.metrics import dp
from kivy.uix.label import Label
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout

from jocular.component import Component
from jocular.widgets.widgets import TextInputC, JMDToggleButton
from jocular.panel import Panel


def exp_to_str(e):
    if e < .001:
        return f'{e*1e6:.0f} µs'
    if e < 1:
        return f'{e*1000:.0f} ms'
    if e - int(e) < .0001:
        return f'{e:.0f}s'
    return f'{e:.2f}s'


def str_to_exp(s):
    s = str(s)
    s = s.lower().strip()
    if len(s) == 0:
        raise Exception
    try:
        if s.endswith('µs') or s.endswith('us'):
            return float(s[:-2].strip()) / 1e6
        if s.endswith('ms'):
            return float(s[:-2].strip()) / 1000
        if s.endswith('s'):
            return float(s[:-1].strip())
        return float(s)
    except:
        raise Exception


class ExposureChooser(Panel, Component):


    def __init__(self, **args):
        super().__init__(**args)
        self.app = App.get_running_app()
        try:
            with open(self.app.get_path('user_exposures.json'), 'r') as f:
                self.user_expos = json.load(f)
        except:
            self.user_expos = []
        self.build()


    def save(self, dt=None):
        with open(self.app.get_path('user_exposures.json'), 'w') as f:
            json.dump(self.user_expos, f, indent=1)       


    def on_show(self):
        # ensure exposurechooser reflects current state
        e_str = exp_to_str(Component.get('CaptureScript').get_exposure())
        for e, but in self.expo_buttons.items():
            but.state = 'down' if e == e_str else 'normal'


    def _button(self, name, expo):
        return JMDToggleButton(
                text=name, 
                size_hint=(.2, None),
                group='expos',
                font_size='18sp',
                height=dp(24),
                on_press=partial(self.exposure_selected, expo))


    def build(self, *args):

        self.expos = [
            .001, .002, .003, .005, .01, .02, .05, .1, .2, .5, 
            1, 2, 5, 10, 15, 20, 25, 30, 45, 60
        ]

        content = self.contents
        self.title_label = Label(text='Select exposure', font_size='24sp')
        self.header.add_widget(self.title_label)

        self.expo_buttons = {}
        for e in self.expos + self.user_expos:
            e_str = exp_to_str(e)
            self.expo_buttons[e_str] = self._button(e_str, e)

        # spacer
        content.add_widget(Label(text='', size_hint=(1, 1)))

        # short exposures
        content.add_widget(Label(
            text='short', 
            size_hint=(1, None), 
            font_size='20sp',
            height=dp(30)))
        gl = GridLayout(
            size_hint=(1, None), 
            cols=5,
            height=dp(120),
            spacing=(dp(5), dp(5)))
        for e in [i for i in self.expos if i < 1]:
            gl.add_widget(self.expo_buttons[exp_to_str(e)])
        content.add_widget(gl)

        # longer exposures
        content.add_widget(Label(
            text='long', 
            size_hint=(1, None), 
            font_size='20sp',
            height=dp(30)))
        gl = GridLayout(
            size_hint=(1, None), 
            cols=5,
            height=dp(120),
            spacing=(dp(5), dp(5)))
        for e in [i for i in self.expos if i >= 1]:
            gl.add_widget(self.expo_buttons[exp_to_str(e)])
        content.add_widget(gl)

        # recently used exposure here; limit to 10 most recent, in order
        content.add_widget(Label(
            text='user-defined', 
            size_hint=(1, None), 
            font_size='20sp',
            height=dp(30)))
        self.user_gl = GridLayout(
            size_hint=(1, None), 
            cols=5,
            height=dp(120),
            spacing=(dp(5), dp(5)))
        for e in sorted(self.user_expos):
            self.user_gl.add_widget(self.expo_buttons[exp_to_str(e)])
        # add some blank labels
        for e in range(10 - len(self.user_expos)):
            self.user_gl.add_widget(Label(size_hint=(.2, None), height=dp(28)))
        content.add_widget(self.user_gl)

        # exposure box
        bv = BoxLayout(
            orientation='vertical',
            size_hint=(1, None),
            height=dp(90))
        bv.add_widget(Label(
            text='add new exposure', 
            size_hint=(1, None),
            font_size='20sp',
            height=dp(32)))
        self.user_exposure_box = TextInputC(
            hint_text='0.3, 10, 10s, 30ms, 50us', 
            size_hint=(.65, None), 
            multiline=False,
            height=dp(30),
            pos_hint={'center_x': .5},
            font_size='20sp',
            on_text_validate=self.custom_exposure_added)
        bv.add_widget(self.user_exposure_box)
        content.add_widget(bv)

        content.add_widget(Label(text='', size_hint=(1, 1)))

        self.app.gui.add_widget(self)


    @logger.catch()
    def custom_exposure_added(self, tb):
        try:
            self.exposure = str_to_exp(self.user_exposure_box.text)
            if self.exposure not in self.expos + self.user_expos:
                self.user_expos += [self.exposure]
            if len(self.user_expos) > 10:
                self.user_expos = self.user_expos[-10:]
            self.user_gl.clear_widgets()
            for e in sorted(self.user_expos):
                self.user_gl.add_widget(self._button(exp_to_str(e), e))
            for e in range(11 - len(self.user_expos)):
                self.user_gl.add_widget(Label(size_hint=(.2, None), height=dp(30)))
            self.user_exposure_box.text = ''
            self.save()
        except:
            pass


    def exposure_selected(self, expo, *args):
        # set exposure on GUI and on scripts panel
        Component.get('CaptureScript').exposure_changed(expo)




''' DeviceFamily:
		superclass of e.g. Camera, Telescope, FilterWheel
		handles communication with devices for generic functions such as
		select, connect, disconnect as well as common error handling

	Device:
		superclass of device instances e.g. SXCamera, ASCOMFilterWheel
'''

import json
import importlib

from kivy.app import App
from loguru import logger
from kivy.event import EventDispatcher

from kivy.properties import (
	ObjectProperty, 
	StringProperty, BooleanProperty, DictProperty
	)
from kivy.clock import Clock

from jocular.component import Component
from jocular.settingsmanager import SettingsBase


class DeviceFamily:

	device = ObjectProperty(None)

	# these three need to be set in each subclass
	family = StringProperty('Unknown')
	modes = DictProperty({})
	default_mode = StringProperty('')


	def __init__(self, **kwargs):
		self.app = App.get_running_app()
		try:
			with open(self.app.get_path(f'{self.family}.json'), 'r') as f:
				self.settings = json.load(f)
		except:
			self.settings = {}
		Clock.schedule_once(self.post_init, 0)


	def post_init(self, dt):
		self.set_mode(self.settings.get('current_mode', self.default_mode))
		self.connect()


	def save(self):
		with open(self.app.get_path(f'{self.family}.json'), 'w') as f:
			json.dump(self.settings, f, indent=1)       


	def set_mode(self, mode):
		''' finds and imports class representing chosen mode
			In future, place all devices in subdirs as in cameras to
			avoid having to import in 'parent' device 
		'''
		self.disconnect()
		try:
			if mode in self.modes:
				devmod = importlib.import_module(f'jocular.{self.family.lower()}')
				devclass = getattr(devmod, self.modes[mode])
				self.device = devclass()
				self.settings['current_mode'] = mode
				self.device.settings_have_changed()
				# self.save()
		except Exception as e:
			logger.exception(e)


	def get_configurables(self):
		if self.device is not None:
			return self.device.configurables


	def configure(self):
		if self.device is not None:
			logger.debug(f"family {self.family} settings {self.settings['current_mode']}")
			self.device.configure()


	def connect(self):
		logger.debug(f"Connecting {self.family} (current mode: {self.settings['current_mode']})")
		if self.device is not None:
			self.device.connect()
			# only save current mode if we are able to connect
			if self.device.connected:
				self.save()
				self.device_connected()
				self.device.on_new_object()


	def disconnect(self):
		if self.device is None:
			return
		if self.connected():
			self.device.disconnect()
			self.device_disconnected()


	def connected(self):
		if not hasattr(self, 'device'):
			return False
		if self.device is None:
			return False
		return self.device.connected


	def device_connected(self):
		pass


	def device_disconnected(self):
		pass


	def on_close(self, *args):
		if self.connected():
			self.disconnect()


	def choose(self, *args):
		if self.device is not None:
			self.device.choose()


''' Each actual device e.g. ASCOMTelescope, ManualFilterwheel etc is a subclass of this
''' 

class Device(EventDispatcher, SettingsBase):

	connected = BooleanProperty(False)
	status = StringProperty('')
	family = StringProperty('')


	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		''' register this device with the device manager e.g. so that it
			appears on the chooser
		'''
		if self.family:
			Component.get('DeviceManager').register(self, name=self.family)


	def on_close(self):
		pass


	def on_new_object(self):
		pass


	def on_previous_object(self):
		pass


	def connect(self):
		self.status = f'Not implemented for this {self.family}'
		self.connected = False


	def disconnect(self):
		self.status = 'not connected'
		self.connected = False


	def on_connected(self, *args):
		Component.get('DeviceManager').connection_changed(self.family, self.connected)


	def on_status(self, *args):
		Component.get('DeviceManager').status_changed(self.family, self.status)


	def select(self, f):
		return None


	def choose(self):
		pass


	def handle_failure(self, message='problem'):
		logger.error(f'{self.family}: failure {message}')
		self.disconnect()
		self.connected = False
		self.status = message
		if hasattr(self, 'on_failure') and self.on_failure is not None:
			self.on_failure()




''' Astrometry utilities. Will support future extension to charting.
'''

import math
import numpy as np
from scipy.special import cotdg

def to360(v):
    if type(v) == list or type(v) == np.ndarray:
        v[v > 360] = v[v > 360] - 360
        v[v < 0] = v[v < 0] + 360
    else:
        if v > 360:
            v -= 360
        elif v < 0:
            v += 360
    return v


def angle360(angle):
    if angle < 0:
        return angle + 360
    elif angle > 360:
        return angle - 360
    else:
        return angle

def angle_diff(t1, t2):
    # anticlockwise angle between 2 positions relative to 0 = positive x-axis
    return angle360(math.atan2(t2[1] - t1[1], t2[0] - t1[0]) / (math.pi / 180))


def spherical_distance(ra, dec, ra0, dec0):
    # return distances of points in ra, dec arrays to point (ra0, dec0)
    d1 = np.sin(np.abs(np.radians(dec) - np.radians(dec0)) / 2) ** 2
    d2 = np.cos(np.radians(dec0)) * np.cos(np.radians(dec)) * (np.sin(np.abs(np.radians(ra) - np.radians(ra0)) / 2) ** 2)
    return np.degrees(2 * np.arcsin((d1 + d2) ** .5))


def Eq2Cart(ra, dec, ra0, dec0):
    # conversion from RA/Dec to standard coords; Snyder (1987) ch 22
    # from matlab version; not fully tested

    rads = np.pi / 180.
    dec = rads * dec
    ra_norm = rads * (ra - ra0)

    if dec0 == 90: 
        # north or south pole so use polar gnomic
        x = np.cos(dec) * np.sin(ra_norm)
        y = -cotdg(dec / rads) * np.cos(ra_norm)

    elif dec0 == -90:
        x = -np.cos(dec) * np.sin(ra_norm)
        y = cotdg(dec / rads) * np.cos(ra_norm)
        
    else:
        # oblique gnomic
        dec0 = rads * dec0
        cos0 = np.cos(dec0)
        sin0 = np.sin(dec0)
        cosdec = np.cos(dec)
        sindec = np.sin(dec)
        cos_ra_norm = np.cos(ra_norm)
        den = cos0 * cosdec * cos_ra_norm + sin0 * sindec
        x = (cosdec * np.sin(ra_norm)) / den
        y = (cos0 * sindec - sin0 * cosdec * cos_ra_norm ) / den

    return -x, y  # reverse RA axis

def Cart2Eq(x, y, ra0, dec0):
    # convert from standard coords to RA/Dec (Snyder ch22)
        
    rads = np.pi / 180.

    # unreverse 'ra'
    x = -x
    rho = (x**2 + y**2)**.5
    c = np.arctan(rho)
    cosc, sinc = np.cos(c), np.sin(c)
    
    if dec0 == 90:
        dec = np.arcsin(cosc)
        ra = np.arctan2(x, -y)
        
    elif dec0 == -90:
        dec = np.arcsin(-cosc)
        ra = np.arctan2(x, y)
        
    else:
        cos0, sin0 = np.cos(rads * dec0), np.sin(rads * dec0)
        dec = np.arcsin(cosc * sin0 + ((y * sinc * cos0 / rho)))
        ra = np.arctan2(x * sinc, rho * cosc * cos0 - y * sinc * sin0)
        
    dec /= rads
    ra = ra0 + ra / rads
    dec[rho==0] = dec0
    ra[rho==0] = ra0
    ra[ra > 360] -= 360
    ra[ra < 0] += 360
    return ra, dec


''' tile represent fields of view
''' 

def intstep(d, step):
    return int(np.floor(step * np.floor(d / step)))


def make_tile(ra0, dec0, fov=3):
    ''' A tile represents all pertinent information about a field
        centred on ra0, dec0 with given fov
    '''
    rads = np.pi / 180.
    fov2 = fov / 2
    polar = ((dec0 + fov2) >= 90) | ((dec0 - fov2) <= -90)
    northern = dec0 >= 0
    if polar:
        min_ra, max_ra = 0, 360
        if northern:
            min_dec, max_dec = 90 - fov2, 90
        else:
            min_dec, max_dec = -90, -90 + fov2
    else:
        ra_width = fov2 / np.cos(dec0 * rads)
        min_ra, max_ra =  to360(ra0 - ra_width), to360(ra0 + ra_width)
        min_dec, max_dec = dec0 - fov2, dec0 + fov2
    return {
        'fov': fov, 'min_ra': min_ra, 'max_ra': max_ra,
        'min_dec': min_dec, 'max_dec': max_dec, 'ra0': ra0, 'dec0': dec0,
        'polar': polar, 'northern': northern
    }

def get_tiles(tile, rstep, dstep):
    ''' return database tiles covering region defined in tile
    '''
    rmin, rmax = tile['min_ra'], tile['max_ra']
    dmin, dmax = tile['min_dec'], tile['max_dec']

    dec_tiles = np.arange(intstep(dmin, dstep), intstep(dmax, dstep) + dstep, dstep)
    if rmin < rmax:
        ra_tiles = np.arange(intstep(rmin, rstep), intstep(rmax, rstep) + rstep, rstep)
    else:
        # handle wraparound
        ra_tiles = \
            list(range(intstep(rmin, rstep), 360, rstep)) + \
            list(range(0, intstep(rmax, rstep) + rstep, rstep))
    return ra_tiles, dec_tiles        

def in_tile(tile, ras, decs):
    ''' return indices of coordinates within tile
    '''
    rmin, rmax = tile['min_ra'], tile['max_ra']
    dmin, dmax = tile['min_dec'], tile['max_dec']
    if rmin < rmax:
        return (decs >= dmin) & (decs < dmax) & (ras >= rmin) & (ras < rmax)
    else:
        return (decs >= dmin) & (decs < dmax) & ((ras >= rmin) | (ras < rmax))


def radec2pix(ra, dec, ra0, dec0, projection):
    ''' return pixel coordinates corresponding to ra and dec,
        referenced to centre of closest matching tile
    '''
    if type(ra) != list and type(ra) != np.ndarray:
        ra = np.array([ra])
        dec = np.array([dec])
    x, y = Eq2Cart(ra, dec, ra0, dec0)
    d = projection(np.array([x, y]).T)
    return d[:, 0], d[:, 1]


def pix2radec(x, y, ra0, dec0, projection):
    ''' find RA/Dec at pixel value
    '''
    if type(x) != list and type(x) != np.ndarray:
        x = np.array([x])
        y = np.array([y])
    d = projection.inverse(np.array([x, y]).T)
    x, y = Cart2Eq(d[:, 0], d[:, 1], ra0, dec0)
    if len(x) == 1:
        return x[0], y[0]
    return x, y




import math

from kivy.lang import Builder
from kivy.uix.scatter import Scatter

from jocular.uranography import angle360
from jocular.component import Component
from jocular.metrics import Metrics


Builder.load_string('''

<ScatterView>:
    size: image.size
    size_hint: None, None 
    auto_bring_to_front: False
    do_rotation: True
    do_translation: True
    do_scale: True
    Image:
        id: image
''')


class ScatterView(Scatter):
    ''' A class representing the moveable image displayed in
        the eyepiece
    '''

    def __init__(self, controller, **kwargs):
        super().__init__(**kwargs)
        self.controller = controller


    def on_touch_down(self, touch):
        # check if touch is within eyepiece and if so, where

        x, y = touch.pos
        xc, yc = Metrics.get('origin')
        r = (((xc - x) ** 2 + (yc - y) ** 2) ** .5) / Metrics.get('inner_radius')

        # touch is not in the eyepiece
        if r  >  .98:
            return False

        # in image
        in_image = self.collide_point(*touch.pos)

        # is it in the outer zone of the image, or not touching image, rotate
        if not in_image or r > .8:
            self.controller.ring_selected = True
            self.theta = math.atan2(y - yc, x - xc) / (math.pi/ 180)
            self.xc = xc
            self.yc = yc
            return True

        if in_image:
            if touch.is_double_tap:
                self.controller.invert = not self.controller.invert
                return True
            self.controller.image_selected  = True
            self.last_x = x
            self.last_y = y
            return True

        return False


    def on_touch_move(self, touch):
        if self.controller.ring_selected:
            x, y = touch.pos
            theta = math.atan2(y - self.yc, x - self.xc) / (math.pi / 180)
            self.controller.orientation = angle360(self.controller.orientation + (theta - self.theta))
            self.theta = theta
            return True
        if self.controller.image_selected:
            x, y = touch.pos
            delta_x = (x - self.last_x)
            delta_y = (y - self.last_y)
            self.x += delta_x
            self.y += delta_y
            self.last_x = x
            self.last_y = y
            Component.get('Annotator').update()
            return True
        return False


    def on_touch_up(self, touch):
        self.controller.ring_selected = False
        self.controller.image_selected = False




''' see https://github.com/python-zwoasi/python-zwoasi/blob/master/zwoasi/__init__.py
'''

import time
import os
import numpy as np
import zwoasi as asi
from loguru import logger

from kivy.app import App
from kivy.properties import StringProperty, NumericProperty, BooleanProperty
from kivy.clock import Clock

from jocular.cameras.genericcamera import GenericCamera
from jocular.utils import toast


class ASICamera(GenericCamera):

	gain = NumericProperty(150)
	offset = NumericProperty(1)
	binning = StringProperty('1 x 1')
	use_min_bandwidth = BooleanProperty(True)
	polling_interval = NumericProperty(.2)  # in seconds

	configurables = [
		('gain', {'name': 'gain setting', 'float': (0, 1000, 10),
				'help': ''}),
		('offset', {'name': 'offset', 'float': (1, 240, 1),
				'help': ''}),
		('binning', {
			'name': 'bin', 
			'options': ['1 x 1', '2 x 2', '3 x 3'],
			'help': ''}),
		('polling_interval', {
			'name': 'polling interval', 
			'float': (.05, 1, .05),
			'help': 'how often to check for exposure'}),
		('use_min_bandwidth', {
			'name': 'minimise USB bandwidth',
			'switch': ''
			})
	]


	def _set_configurable(self, prop, vals):
		self.configurables = [(k, vals if k==prop else v) for k, v in self.configurables]


	def connect(self):

		if self.connected:
			return
			
		self.connected = False
		self.asicamera = None

		try:
			asipath = App.get_running_app().get_path('ASI')
			if not os.path.exists(asipath):
				self.status = 'Cannot find ASI library in resources'
				return
		except Exception as e:
			self.status = 'cannot find ASI library'
			logger.exception('asipath problem ({:})'.format(e))
			return

		try:
			asi.init(asipath)
		except asi.ZWO_Error as e:
			# we don't mind if it is already initialised but
			# we care about other errors
			if str(e).strip() == 'Library already initialized':
				pass
			else:
				self.status = 'cannot initialise camera'
				logger.exception('problem initialising ASI ({:})'.format(e))
				return
		except Exception as e:
			self.status = 'cannot initialise camera'
			logger.exception('problem initialising ASI ({:})'.format(e))
			return

		try:
			num_cameras = asi.get_num_cameras()
			if num_cameras == 0:
				self.status = 'no camera found'
				return
		except Exception as e:
			self.status = 'cannot get number of cameras'
			logger.exception('get_num_cameras ({:})'.format(e))
			return

		# get first camera found
		try:
			self.asicamera = asi.Camera(0)
		except Exception as e:
			self.status = 'error assigning camera 0'
			logger.exception('problem assigning camera 0 ({:})'.format(e))
			return

		# examine camera properties
		try:
			self.camera_props = self.asicamera.get_camera_property()
			for k, v in self.camera_props.items():
				logger.info('{:15s} {:}'.format(k, v))
		except Exception as e:
			self.status = 'error getting camera properties'
			logger.exception('error getting camera properties ({:})'.format(e))
			return

		# list camera controls
		try:
			self.camera_controls = self.asicamera.get_controls()
			for k, v in self.camera_controls.items():
				logger.info('{:15s} {:}'.format(k, v))
		except Exception as e:
			self.status = 'error getting camera controls'
			logger.exception('error getting camera controls ({:})'.format(e))
			return

		# get bin options
		try:
			binops = ['{:1d} x {:1d}'.format(b, b) for b in self.camera_props['SupportedBins']]
			self._set_configurable('binning', {
				'name': 'binning', 
				'options': binops})
		except Exception as e:
			logger.warning('cannot get bin options ({:})'.format(e))
			binops = ['1 x 1']

		# get gain options
		try:
			gainprops = self.camera_controls['Gain']
			ming, maxg = gainprops['MinValue'], gainprops['MaxValue']
			self._set_configurable('gain', {
				'name': 'gain setting', 
				'float': (ming, maxg, 10)})
		except Exception as e:
			self.status = 'cannot get gain options'
			logger.exception('cannot get gain options ({:})'.format(e))
			return

		# get offset options
		try:
			offsetprops = self.camera_controls['Offset']
			ming, maxg = offsetprops['MinValue'], offsetprops['MaxValue']
			self._set_configurable('offset', {
				'name': 'offset', 
				'float': (ming, maxg, 1)})
		except Exception as e:
			self.status = 'cannot get gain options'
			logger.exception('cannot get gain options ({:})'.format(e))
			return

		# set some properties
		try:
			if self.use_min_bandwidth:
				self.asicamera.set_control_value(asi.ASI_BANDWIDTHOVERLOAD, 
					self.camera_controls['BandWidth']['MinValue'])
			self.asicamera.disable_dark_subtract()
			self.asicamera.set_image_type(asi.ASI_IMG_RAW16)
		except Exception as e:
			self.status = 'error setting camera values'
			logger.exception('error setting camera values ({:})'.format(e))
			return

		try:
			# whbi = self.asicamera.get_roi_format()
			#if 'temperature' in self.camera_controls:
			temp = self.asicamera.get_control_value(asi.ASI_TEMPERATURE)[0]
			#else:
			#	temp = None
			self.status = '{:}: {:} x {:} {:}'.format(
				self.camera_props.get('Name', 'anon'),
				self.camera_props.get('MaxWidth', 0),
				self.camera_props.get('MaxHeight', 0),
				'' if temp is None else '{:.0f}C'.format(temp / 10))
				
			self.connected = True
			logger.info(self.status)
		except Exception as e:
			self.status = 'error getting camera sensor size'
			logger.exception('error getting camera sensor size ({:})'.format(e))

		# set initial ROI, binning etc
		self.on_gain()
		self.on_offset()
		self.set_ROI(ROI=None)


	def capture(self, exposure=None, on_capture=None, on_failure=None, is_faf=False,
		binning=None, return_image=False, is_bias=False):

		if is_bias:
			self.exposure = int(self.camera_controls['Exposure']['MinValue']) / 1e6
		else:
			self.exposure = exposure

		self.on_failure = on_failure
		self.on_capture = on_capture

		# cancel any existing exposure
		try:
			self.asicamera.stop_exposure()
		except Exception as e:
			toast('error stopping exposure')
			logger.exception('error stopping exposure ({:})'.format(e))

		# exposure is in microseconds (this is fast to set)
		try:
			self.asicamera.set_control_value(
				asi.ASI_EXPOSURE, int(self.exposure * 1e6))
			# logger.info('setting exposure to {:}'.format(self.exposure))
		except Exception as e:
			toast('camera issue while setting exposure')
			logger.exception('camera issue while setting exposure ({:})'.format(e))

		self.asicamera.start_exposure()

		if return_image:
			time.sleep(self.exposure)
			ready = False
			while not ready:
				time.sleep(self.polling_interval)
				ready = self.asicamera.get_exposure_status() != asi.ASI_EXP_WORKING
			if self.asicamera.get_exposure_status() != asi.ASI_EXP_SUCCESS:
				self.handle_failure('capture problem')
				return
			cd = self.get_camera_data()
			if cd is None:
				self.handle_failure('capture problem')
			return cd
			 
		self.capture_event = Clock.schedule_once(self.check_exposure, 
			max(self.polling_interval, self.exposure))


	def set_ROI(self, ROI=None):
		''' ROI has changed
		'''
		if not self.connected:
			return

		bins=int(self.binning[0])

		# if ROI is unclicked, use settings on interface
		if ROI is None:
			self.asicamera.set_roi(bins=bins)
			logger.trace('ROI changed: to full sensor, bin {:} '.format(bins))
			return

		# use ROI, which assumes binning already applied during framing
		start_x, width, start_y, height = ROI
		width = 8 * (width // 8)   # must be mult of 8
		height = 2 * (height // 2) # and of 2

		# note that width is a mult of 8 and height is mult of 2
		self.asicamera.set_roi(
			bins=bins,
			start_x=start_x,
			start_y=start_y,
			width=width,
			height=height)

		logger.trace('ROI changed: bins {:} start_x {:} start_y {:} width {:} height {:}'.format(
			bins, start_x, start_y, width, height))


	def on_binning(self, *args):
		''' If user changes binning then ROI is reset to None
			for the moment at least
		'''
		self.set_ROI(ROI=None)

		
	def on_gain(self, *args):
		if not self.connected:
			return
		try:
			self.asicamera.set_control_value(asi.ASI_GAIN, int(self.gain))
			logger.info('setting gain to {:.0f}'.format(self.gain))
		except Exception as e:
			toast('problem setting gain')
			logger.exception('problem setting gain ({:})'.format(e))


	def on_offset(self, *args):
		if not self.connected:
			return
		try:
			''' apparently offset is called brightness
			'''  
			self.asicamera.set_control_value(asi.ASI_BRIGHTNESS, int(self.offset))
			logger.info('setting offset to {:.0f}'.format(self.offset))
		except Exception as e:
			toast('problem setting offset')
			logger.exception('problem setting offset ({:})'.format(e))

	def get_camera_data(self):
		''' see capture method in
			https://github.com/stevemarple/python-zwoasi/blob/master/zwoasi/__init__.py
		'''

		# sometimes gets a timeout for reasons that are not clear
		try:
			data = self.asicamera.get_data_after_exposure(None)
			whbi = self.asicamera.get_roi_format()
			shape = [whbi[1], whbi[0]]
			img = np.frombuffer(data, dtype=np.uint16)
			return img.reshape(shape) / 2 ** 16
		except Exception as e:
			logger.warning('{:}'.format(e))
			return None


	def get_capture_props(self):
		''' return dict of props such as gain, ROI etc
			(these get converted to uppercase in FITs)
		'''
		start_x, start_y, width, height = self.asicamera.get_roi()
		return {
			'camera': self.camera_props.get('Name', 'anon'),
			'gain': self.gain,
			'offset': self.offset,
			'binning': int(self.binning[0]),
			'temperature': self.get_sensor_temperature(),
			'ROI_x': start_x,
			'ROI_y': start_y,
			'ROI_w': width,
			'ROI_h': height,
			'pixel_width': self.camera_props['PixelSize'], # this before binning I think
			'pixel_height': self.camera_props['PixelSize']
			}


	def get_pixel_height(self):
		try:
			return self.camera_props['PixelSize'] * int(self.binning[0])
		except:
			return None


	def get_sensor_temperature(self):
		try:
			return self.asicamera.get_control_value(asi.ASI_TEMPERATURE)[0] / 10
		except:
			return None


	def check_exposure(self, *arg):
		# if ready, get image, otherwise schedule next check 
		if self.asicamera.get_exposure_status() != asi.ASI_EXP_WORKING:
			self.last_capture = self.get_camera_data()
			if self.last_capture is not None:
				# got data, so schedule callback
				if self.on_capture is not None:
					Clock.schedule_once(self.on_capture, 0)
		else:
			self.capture_event = Clock.schedule_once(
				self.check_exposure, 
				self.polling_interval)


	def stop_capture(self):
		# Cancel any pending reads
		if hasattr(self, 'asicamera') and self.asicamera is not None:
			self.asicamera.stop_exposure()
		if hasattr(self, 'capture_event'):
			self.capture_event.cancel()




''' Allows user to select one or more filters to apply
'''

import math
import numpy as np
import json
from functools import partial
from loguru import logger

from kivy.app import App
from kivy.metrics import dp
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.label import Label
from kivymd.uix.slider import MDSlider
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.lang import Builder
from kivy.properties import ListProperty

from jocular.component import Component
from jocular.widgets.widgets import LabelR
from jocular.panel import Panel


Builder.load_string('''

<FilterToggle>:
    canvas.before:
        Color:
            rgba: .15, .15, .15, 1
        Ellipse:
            pos: self.pos
            size: self.width, self.height
        Color:
            rgb: self.filter_color if self.state == 'down' else (.8, .8, .8)
            a: .7 if self.state == 'down' else .2
        Ellipse:
            pos: self.x + 2, self.y + 2
            size: self.width - 5, self.height - 5
    markup: True
    color: (.2, .2, .2, 1) if self.state == 'down' else (.6, .6, .6, 1)
    disabled: self.text == '-'
    background_color: 0, 0, 0, 0
    size: dp(70), dp(70)
    font_size: '20sp' if self.state == 'down' else '16sp'
    size_hint: None, None
''')


class FilterToggle(ToggleButton):
  filter_color = ListProperty([1, 1, 1, 1])


class FilterChooser(Panel, Component):


    def __init__(self, **args):
        super().__init__(**args)
        self.app = App.get_running_app()
        try:
            with open(self.app.get_path('filter_properties.json'), 'r') as f:
                self.filter_properties = json.load(f)
        except Exception as e:
            logger.error(e)
            self.filter_properties = {}
        self.nsubs = 4
        self.build()


    def on_show(self):
        # ensure filterchooser reflects current filterwheel device
        state = Component.get('FilterWheel').get_state()
        logger.debug(f'state from FW {state}')
        logger.debug(f'previous state {self.state}')
        if state['filtermap'] != self.state['filtermap']:
            logger.debug('rebuilding FilterChooser')
            # need to rebuild
            self.state = state
            self.header.clear_widgets()
            self.contents.clear_widgets()
            self.build_fw()
        self.update_panel()


    def build(self, dt=None):
        self.state = Component.get('FilterWheel').get_state()
        if len(self.state['filtermap']) > 0:
            self.build_fw()
        self.app.gui.add_widget(self)


    def build_fw(self, dt=None):
        ''' Generate filterwheel selection widget
        '''

        self.title_label = Label(text='Select filter', font_size='24sp')
        self.header.add_widget(self.title_label)
 
        # filter wheel
        pos2filter = {int(p):f for f, p in self.state['filtermap'].items()}
        n_positions = 9
        angs = np.linspace(0, 2*math.pi, n_positions + 1) # [:-1]
        fw_layout = FloatLayout(
            size_hint=(None, None), 
            size=(dp(300), dp(300)),
            pos_hint={'center_x': .5, 'center_y': .5})
        self.buts = {}
        for i in range(n_positions):
            pos_hint={ 
                'center_x': ( .8 * math.cos(angs[i]) + 1 ) / 2, 
                'center_y': ( .8 * math.sin(angs[i]) + 1 ) / 2
                }
            pos = i + 1
            if pos in pos2filter:
                f = pos2filter[pos]
                b = self.buts[f] = FilterToggle(
                        text=f, filter_color=self.filter_properties[f]['color'], 
                        on_press=partial(self.filter_selected, f),
                        pos_hint=pos_hint)
            else:
                b = FilterToggle(text='-', filter_color=(0, 0, 0, 1), 
                        pos_hint=pos_hint, disabled=True)
            fw_layout.add_widget(Label(text=str(i+1),
                pos_hint={ 
                'center_x': ( .48 * math.cos(angs[i]) + 1 ) / 2, 
                'center_y': ( .48 * math.sin(angs[i]) + 1 ) / 2
                }))
            fw_layout.add_widget(b)

        # nsubs slider
        bh = self.nsubs_box = BoxLayout(size_hint=(1, 1))
        self.nsubs_label = LabelR(
            text='subs/filter', 
            font_size='20sp',
            size_hint=(.3, 1))
        bh.add_widget(self.nsubs_label)
        self.nsubs_slider = MDSlider(min=1, max=20, value=4, step=1, size_hint=(.7, 1))
        self.nsubs_slider.bind(value=self.nsubs_changed)
        bh.add_widget(self.nsubs_slider)

        # build widget
        self.contents.add_widget(Label(size_hint=(1, 1), text=''))
        self.contents.add_widget(fw_layout)
        self.contents.add_widget(bh)


    def on_leave(self, *args):
        # overrides Panel on_leave
        if not hasattr(self, 'buts'):
            return
        filts = [f for f, b in self.buts.items() if b.state == 'down']
        if len(filts) > 0:
            Component.get('CaptureScript').filter_changed(filts, nsubs=self.nsubs)
        self.hide()


    def filter_selected(self, filt, but):
        # set exposure on GUI and on scripts panel
        if Component.get('CaptureScript').current_script == 'seq':
            logger.debug(f'toggled filter {filt}')
        else:
            logger.debug(f'selected filter {filt}')
            for f, b in self.buts.items():
                b.state = 'down' if f == filt else 'normal'
            Component.get('CaptureScript').filter_changed([filt])


    def update_panel(self):
        filters = Component.get('CaptureScript').get_filters()
        for f, but in self.buts.items():
            but.state = 'down' if f in filters else 'normal'
        nsubs = Component.get('CaptureScript').get_nsubs()
        if nsubs > 1:
            self.nsubs_slider.value = nsubs
            self.nsubs_label.text = f'{nsubs} subs/filter'
            self.nsubs = nsubs
            self.title_label.text = 'Select filters'
            self.nsubs_box.disabled = False
        else:
            self.title_label.text = 'Select a filter'
            self.nsubs_box.disabled = True


    def nsubs_changed(self, slider, nsubs):
        self.nsubs_label.text = f'{nsubs} subs/filter'
        self.nsubs = nsubs


    def order_by_transmission(self, filts):
        if type(filts) != list:
            filts = [filts]
        trans = {f: self.filter_properties[f]['trans'] for f in filts}
        return sorted(trans, key=trans.get)


    def get_filter_types(self):
        return list(self.filter_properties.keys())


    def on_touch_down(self, touch):
        handled = super().on_touch_down(touch)
        if self.collide_point(*touch.pos):
            return True
        return handled



''' Support for Jocular-specific icons
'''

joc_icons = {
    'pause': '3',
    'play': 'P', # was 'n',
    'oculus': 'o',
    'last_sub': 'f',
    'first_sub': 'e',
    'next_sub': '>',
    'prev_sub': '<',
    'camera': 'b',
    'reticle': 'r', # was v
    'ROI': 'd', # was 'v'
    'fit': 'y',
    'realign': 'E',
    'clear': '0',
    'quit': 'Q',
    'settings': 'x',
    'snapshotter': 'o',
    'prev': 'F',
    'list': 'w',
    'lever': '1',
    'dot': '1',
    'redo': 'h',  # was '4'
    'new': 'i',
    'save': 'i',  # until we update
    'warn': '!',
    'error': 'W',
    'solve': 'T',
    'slew': 'H',
    'dso': 'a',
    'equip': 'H',
    'show_subs': '}', # was '*'
    'stack': '5',
    'snapshot': 'z',
    'experimental': 'c',
    'info': '}',
    'tooltips': '?'
}


def get_icon(icon_name, font_size=None, color=None):
    ''' Return a Kivy-markup icon
    '''
    font_size = f'{font_size}sp' if font_size else '17sp'
    icon = joc_icons.get(icon_name, icon_name)
    if color is None:
        return f'[font=Jocular][size={font_size}]{icon}[/size][/font]'
    color = {'r': 'ff0000', 'y': 'ffff00', 'g': '00ff00'}[color]
    return f'[font=Jocular][size={font_size}][color={color}]{icon}[/color][/size][/font]'
 


''' Miscellany
'''

import os
import numpy as np
import math
import shutil
from scipy.stats import trimboth
from loguru import logger 
from pathlib import Path
from kivymd.toast.kivytoast import toast as mdtoast


def is_null(v):
    return (v is None) or \
        (isinstance(v, str) and v.strip() == '') or  \
        (isinstance(v, float) and math.isnan(v)) or  \
        (isinstance(v, dict) and len(v) == 0) or \
        (isinstance(v, list) and len(v) == 0)


def toast(mesg, duration=1):
    mdtoast(mesg, duration=duration)
    logger.info(f'toast: {mesg}')


# def angle360(angle):
#     if angle < 0:
#         return angle + 360
#     elif angle > 360:
#         return angle - 360
#     else:
#         return angle


def percentile_clip(a, perc=80):
    return np.mean(trimboth(np.sort(a, axis=0), (100 - perc)/100, axis=0), axis=0)    


def unique_member(l):
    # if list has a single unique member, return it, otherwise None
    if len(set(l)) == 1:
        return l[0]
    else:
        return None


def get_datadir():
    ''' Read datadir name from .jocular in home; if .jocular doesn't exist,
        or if the datadir doesn't exist, caller will deal with it
    '''
    try:
        with open(os.path.join(str(Path.home()), '.jocular'), 'r') as f:
            datadir = f.read().strip()
        if os.path.exists(datadir):
            return datadir
        return None
    except:
        return None


def add_if_not_exists(path):
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except FileExistsError:
            raise FileExistsError


def start_logging(path):
    ''' start logging on given path
    '''

    logger.remove()
    fmt = "{time: DMMMYY HH:mm:ss.SSS} | {level: <8} | {name: <25} | {function: <26} | {line: >4} | {message}"
    logger.add(os.path.join(path, 'jocular.log'), mode='a', format=fmt, level='TRACE',
        rotation='1MB', retention=20)
    for i in range(10):
        logger.info('')
    logger.info('--------------------------')
    logger.info('Started logging')
    logger.info('')


def make_unique_filename(path):
    if not os.path.exists(path):
        return path
    root, name = os.path.split(path) 
    base, ext = os.path.splitext(name)
    cnt = 1
    while os.path.exists(os.path.join(root, f'{base}_{cnt}{ext}')):
        cnt += 1
    return os.path.join(os.path.join(root, f'{base}_{cnt}{ext}'))


def purify_name(nm):
    # replace any chars that could cause problems with filesystem by spaces
    return nm.replace(':',' ').replace('/',' ').replace('\\',' ').strip()

# consider combining with above; called in ObjectIO
def generate_observation_name(path, prefix=None):

    # use names of form obs 1, obs 2 if prefix is obs
    if (prefix is None) or (prefix == 'light') or (len(prefix) == 0):
        n = 1
        while os.path.exists(os.path.join(path, f'obs {n}')):
            n += 1
        return f'obs {n}'

    # otherwise using name is possible
    if not os.path.exists(os.path.join(path, prefix)):
        return prefix

    # otherwise add a version number
    n = 1
    while os.path.exists(os.path.join(path, f'{prefix} v{n}')):
        n += 1
    return f'{prefix} v{n}'

def s_to_minsec(s):
    if s < 1:
        return f'{1000 * s:.0f}ms'
    mins, secs = divmod(round(s), 60)
    if mins == 0:
        return f'{secs}s'
    elif secs == 0:
        return f'{mins}m'
    else:
        return f'{mins:d}m{secs:d}s'

def move_to_dir(frompath, topath):
    ''' move from frompath to topath, making topath dir and 
        creating unique name if necessary
    '''

    try:
        toparent, tobase = os.path.split(topath)
        fromparent, filename = os.path.split(frompath)

        # empty parent implies use fromparent
        if not toparent:
            # treat as subdirectory of from
            topath = os.path.join(fromparent, tobase)
        add_if_not_exists(topath)

        # create unique filename
        dest = make_unique_filename(os.path.join(topath, filename))

        # instead of this
        # os.rename(frompath, dest)    
        # use this to handle case of moving from one filesys to another
        shutil.move(frompath, dest)    

    except Exception as e:
        logger.exception(f'problem moving {frompath} to {topath} ({e})')





import glob
import os
import time
from functools import partial
import numpy as np

from loguru import logger

from kivy.properties import StringProperty
from kivy.clock import Clock
from jocular.cameras.genericcamera import GenericCamera
from jocular.image import Image

#-------- Simulator Camera

class SimulatorCamera(GenericCamera):

	source = StringProperty('noise')

	configurables = [
		('source', {'name': 'How to generate', 'options': ['noise', 'capture'],
				'help': 'use noise or choose a random capture'})
	]

	def on_new_object(self):

		logger.info('Simulator on new object called')

		self.last_capture = None
		self.ims = None
		self.ROI = None

		''' choose capture directory at random
		'''
		if self.source == 'capture':
			sessions = glob.glob(os.path.join(self.app.get_path('captures'), '*'))
			if len(sessions) > 0:
				sesh = sessions[np.random.randint(len(sessions))]
				# choose within session
				dsos = glob.glob(os.path.join(sesh, '*'))
				if len(dsos) > 0:
					dso = dsos[np.random.randint(len(dsos))]
					# change this to work in path-independent way
					self.ims = glob.glob(dso+'/*.fit')
					if len(self.ims) == 0:
						self.ims = None
					else:
						logger.info('Using {:} for images'.format(self.ims))

	def connect(self):
		if self.source == 'capture':
			self.status = 'connected to captures'
		else:
			self.status = 'connected to random generator'
		self.connected = True
		self.on_new_object()

	def capture(self, exposure=None, on_capture=None, on_failure=None, is_faf=False,
		binning=None, return_image=False, is_bias=False):
		logger.info('Capturing in simulator expo {:}'.format(exposure))

		self.exposure = exposure
		if return_image:
			time.sleep(exposure)
			return self.get_camera_data()
			 
		self.capture_event = Clock.schedule_once(
			partial(self.check_exposure, on_capture), max(0.01, exposure))

	def get_capture_props(self):
		return {
			'camera': 'simulator',
			'temperature': None
		}

	def check_exposure(self, on_capture, dt):
		# if ready, get image, otherwise schedule a call in 200ms
		self.last_capture = self.get_camera_data()
		if on_capture is not None:
			Clock.schedule_once(on_capture, 0)

	def stop_capture(self):
		# Cancel any pending reads
		if hasattr(self, 'capture_event'):
			self.capture_event.cancel()

	def get_camera_data(self):
		if self.source == 'capture' and self.ims is not None:
			try:
				impath = self.ims[np.random.randint(len(self.ims))]
				logger.info('Serving fits image {:}'.format(impath))
				im = Image(impath)
				if self.ROI is None:
					return im.get_image() 
				xstart, width, ystart, height = self.ROI
				return im.get_image() [ystart: ystart + height, xstart: xstart + width]
			except Exception as e:
				logger.exception(e)
		# in the event of failure to get image, return noise
		return min(1, (self.exposure / 10)) * np.random.random((500, 500))

	def set_ROI(self, ROI):
		'''
		'''
		self.ROI = ROI





''' Watched camera. Monitors a user-specified folder for new FITs images.
'''

import os
import numpy as np
from skimage.transform import rescale, downscale_local_mean

from loguru import logger
from kivy.app import App
from kivy.properties import StringProperty, BooleanProperty
from kivy.clock import Clock

from jocular.component import Component
from jocular.utils import move_to_dir, toast
from jocular.image import Image, ImageNotReadyException, is_fit
from jocular.cameras.genericcamera import GenericCamera

class WatchedCamera(GenericCamera):

	configurables = [
		('watched_dir', {
			'name': 'Watched directory', 
			'filechooser': 'dironly',
			'help': 'Choose a location for Jocular to monitor'}),
		('save_originals', {
			'name': 'save originals?',
			'switch': '',
			'help': 'Keep FITs that arrive in watched dir?'}),
		('controllable', {
			'name': 'control using Jocular?', 
			'switch': '',
			'help': 'Only monitor when capture control is active'}),
		('binning', {
			'name': 'bin',
			'options': ['no', '2 x 2', '3 x 3', '4 x 4'],
			'help': 'bin images that arrive in the watched dir'
			}),
		('bin_method', {
			'name': 'bin method',
			'options': ['mean', 'interpolate'],
			'help': 'simple average or interpolation'
			}),
		('colour_space', {
			'name': 'colour space',
			'options': ['mono', 'GBRG', 'BGGR', 'GRBG', 'RGGB'],
			'help': 'how to interpret pixel matrix'
			}),
		('exposure', {
			'name': 'exposure',
			'options': ['auto', 'from user'],
			'help': 'how to get exposure (auto: FITs/filename)'
			}),
		('filt', {
			'name': 'filter',
			'options': ['auto', 'from user', 'assume L'],
			'help': 'how to get filter information (ignored when debayering)'
			}),
		('sub_type', {
			'name': 'sub type',
			'options': ['auto', 'from user', 'assume light'],
			'help': 'how to get sub type'
			})
	]

	watched_dir = StringProperty(None)
	save_originals = BooleanProperty(True)
	controllable = BooleanProperty(False)
	binning = StringProperty('no')
	colour_space = StringProperty('mono')
	bin_method = StringProperty('mean')
	exposure = StringProperty('auto')
	filt = StringProperty('auto')
	sub_type = StringProperty('auto')
	temperature = StringProperty('auto')

	capturing = BooleanProperty(False)

	def connect(self):

		self.connected = False

		# if no watched dir try to create in data_dir
		if self.watched_dir is None:
			data_dir = App.get_running_app().data_dir
			if data_dir is not None:
				path = os.path.join(data_dir, 'watched')
				if os.path.exists(path):
					self.watched_dir = path
				else:
					logger.debug('creating watched dir {:}'.format(path))
					try:
						os.mkdir(path)
						self.watched_dir = path
					except Exception as e:
						logger.exception('cannot create watched dir ({:})'.format(e))
						
		# if we reach here and watched dir still none, ask user to select
		if self.watched_dir is None:
			self.status = 'Please configure a watched directory'
			self.connected = False
			return

		# check if we can write to watched directory			
		try:
			with open(os.path.join(self.watched_dir, '.written'), 'w') as f:
				f.write('can write')
		except:
			self.status = 'Cannot write to {:}'.format(self.watched_dir)
			self.connected = False
			self.watched_dir is None
			return

		# start watcher
		self.watching_event = Clock.schedule_interval(self.watch, .2)
		self.connected = True
		self.capturing = True  #  make this a setting
		self.status = 'Watching {:}'.format(self.watched_dir)

	def disconnect(self):
		if hasattr(self, 'watching_event'):
			self.watching_event.cancel()
			self.flush()
		self.status = 'stopped monitoring'
		self.connected = False

	def on_new_object(self):
		self.flush()
		self.capturing = not self.controllable

	def on_previous_object(self):
		self.capturing = False

	def capture(self, **kwargs):
		self.capturing = True 
		
	def stop_capture(self, **kwargs):
		if self.controllable: 
			self.capturing = False 

	def flush(self):
		# move any FITs to 'unused' folder in watcher
		for path in self.get_possible_fits():
			move_to_dir(path, 'unused')

	def on_controllable(self, *args):
		self.capturing = not self.controllable

	def get_possible_fits(self):
		wdir = self.watched_dir
		if wdir is None or not os.path.exists(wdir):
			return []
		return [os.path.join(wdir, d) for d in os.listdir(wdir) if is_fit(d)]

	def watch(self, dt):
		''' Monitor watched directory.
		'''

		if self.watched_dir is None or not os.path.exists(self.watched_dir):
			return

		for path in self.get_possible_fits():
			s = None
			if self.capturing:

				try:
					s = Image(path, check_image_data=True)
				except ImageNotReadyException as e:
					# give it another chance on next event cycle
					logger.debug('image not ready {:} ({:})'.format(path, e))
					# pass
				except Exception as e:
					logger.exception('other issue {:} ({:})'.format(path, e))
					toast('Invalid fits file')
					move_to_dir(path, 'invalid')

				if s is not None:
					try:
						self.process_sub(s, path)
					except Exception as e:
						logger.exception('error processing sub ({:})'.format(e))
						toast('error processing sub')
						move_to_dir(path, 'invalid')
			else:				
				move_to_dir(path, 'unused')


	def process_sub(self, s, path):
		''' Apply user-specified colour space conversion and binning. Also 
			check if image has 3 separated RGB planes and generate 3 separate
			subs. 
		'''

		bn = os.path.basename(path)
		im = s.get_image()
		imshape = str(im.shape)
		imstats = 'mean {:.4f}. range {:.4f}-{:.4f}'.format(np.mean(im), np.min(im), np.max(im))

		# check if image has 3 separate (RGB) planes
		if im.ndim == 3:
			logger.debug('sub RGB {:} {:}'.format(imshape, imstats))
			if im.shape[0] == 3:  # 3 x X x Y
				for i, chan in enumerate(['R', 'G', 'B']):
					self.save_mono(s, self.bin(im[i, :, :]), bn, filt=chan, sub_type='light')
			elif im.shape[2] == 3:  # X x Y x 3
				for i, chan in enumerate(['R', 'G', 'B']):
					self.save_mono(s, self.bin(im[:, :, i]), bn, filt=chan, sub_type='light')
			else:
				# shouldn't happen as im nims/shape is checked in Image
				logger.error('cannot process 3D image')

		# don't debayer if mono
		elif self.colour_space == 'mono':
			logger.debug('sub mono {:} {:}'.format(imshape, imstats))
			self.save_mono(s, self.bin(im), bn)

		# don't debayer if master calibration frame
		elif s.is_master:
			logger.debug('master mono {:} {:}'.format(imshape, imstats))
			self.save_mono(s, self.bin(im), bn)

		# debayer
		else:
			logger.debug('sub OSC {:} {:}'.format(imshape, imstats))
			from colour_demosaicing import demosaicing_CFA_Bayer_bilinear
			rgb = demosaicing_CFA_Bayer_bilinear(im, pattern=self.colour_space)
			# rescale to original intensity range
			cfa_min, cfa_max = np.min(im), np.max(im)
			rgb_min, rgb_max = np.min(rgb), np.max(rgb)
			rgb = cfa_min + (cfa_max - cfa_min) * (rgb - rgb_min) / (rgb_max - rgb_min)
			for i, chan in enumerate(['R', 'G', 'B']):
				self.save_mono(s, self.bin(rgb[:, :, i]), bn, filt=chan, sub_type='light')
 
		if self.save_originals:
			Component.get('ObjectIO').save_original(path)
		else:
			Component.get('ObjectIO').delete_file(path)



	def save_mono(self, sub, im, nm, filt=None, sub_type=None):
		''' Save image, binning if required, constructing details from Image instance
		'''

		cs = Component.get('CaptureScript')

		# add header info from sub into capture_props
		# note that this is just the essential info that
		# survives even after jocular has done its
		# processes
		capture_props = { p: getattr(sub, p) if hasattr(sub, p) else None 
			for p in ['gain', 'offset', 'camera', 'pixel_height', 'pixel_width', 'binning']}

		# if jocular has binned, changed pixel dims
		b = self.binning
		if b != 'no':
			binfac = int(b[0])
			if capture_props['pixel_height'] is not None:
				capture_props['pixel_height'] *= binfac
			if capture_props['pixel_width'] is not None:
				capture_props['pixel_width'] *= binfac


		capture_props['exposure'] = {
			'auto': sub.exposure,
			'from user': cs.get_exposure(),
		}[self.exposure]

		capture_props['filter'] = filt 
		if capture_props['filter'] is None:
			''' if from user, choose first filter in case of seq i.e. won't
				really work to use seq with watched
			'''
			capture_props['filter'] = {
				'auto': sub.filter,
				'from user': cs.get_filters()[0],
				'assume L': 'L'
			}[self.filt]

		capture_props['sub_type'] = sub_type
		if capture_props['sub_type'] is None:
			capture_props['sub_type'] = {
				'auto': sub.sub_type,
				'from user': cs.get_sub_type(),
				'assume light': 'light'
			}[self.sub_type]

		capture_props['temperature'] = {
			'auto': sub.temperature,
			'from user': Component.get('Session').temperature,
		}[self.temperature]

		# don't scale masters (check this!)
		if sub.is_master:
			capture_props['nsubs'] = sub.nsubs
			Component.get('Calibrator').save_master(
				data=im,
				capture_props=capture_props)

		else:
			# send image to ObjectIO to save
			# rescale to 16-bit int
			im *= (2**16)
			Component.get('ObjectIO').new_sub(
				data=im.astype(np.uint16),
				name='{:}_{:}'.format(capture_props['filter'], nm),
				capture_props=capture_props)

		# signal to interface the values being used
		cs.set_external_details(
			filt=capture_props['filter'],
			exposure=capture_props['exposure'],
			sub_type=capture_props['sub_type'])

	def bin(self, im):
		b = self.binning
		if b == 'no':
			return im
		binfac = int(b[0]) # i.e. 2 x 2 gives 2, etc
		if self.bin_method == 'interpolation':
			return rescale(im, 1 / binfac, anti_aliasing=True, mode='constant', 
				preserve_range=True, multichannel=False)
		return downscale_local_mean(im, (binfac, binfac))





''' Supports annotation browsing functionality.
'''

import numpy as np

from kivy.app import App
from loguru import logger
from kivy.properties import (
    NumericProperty,
    StringProperty,
    BooleanProperty,
    ListProperty,
    DictProperty,
)
from kivy.uix.label import Label
from kivy.uix.boxlayout import BoxLayout
from kivy.metrics import dp
from kivy.lang import Builder
from kivy.core.window import Window
from kivymd.uix.behaviors import HoverBehavior

from jocular.metrics import Metrics
from jocular.component import Component
from jocular.uranography import make_tile, radec2pix


Builder.load_string(
'''

#:import np numpy

<PropVal>:
    canvas.before:
        Color:
            rgba: .3, .3, .3, .9
        Rectangle:
            pos: self.pos
            size: self.size
    size_hint: 1, None
    height: dp(20)
    Label:
        size_hint: None, 1
        markup: True
        halign: 'left'
        text_size: self.size
        color: [.5, .5, .5, .7] # app.lowlight_color
        font_size: app.form_font_size
        background_color: .6, 0, .6, 0
        width: dp(100)
        padding: dp(5), dp(1)
        text: f'[b]{root.param}[/b]'
    Label:
        size_hint: None, 1
        markup: True
        halign: 'left'
        text_size: self.size
        color: [.5, .5, .5, .7] # app.lowlight_color
        font_size: app.form_font_size
        background_color: .6, 0, .6, 0
        width: dp(250)
        text: root.value


<TitleBar>:
    canvas.before:
        Color:
            rgba: .4, .4, .4, .3
        Rectangle:
            pos: self.pos
            size: self.size
    size_hint: 1, None
    height: dp(30)
    padding: dp(5), dp(2)
    Label:
        size_hint: None, 1
        markup: True
        halign: 'left'
        shorten: True
        shorten_from: 'right'
        valign: 'middle'
        text_size: self.size
        color: root.title_color
        font_size: str(int(app.form_font_size[:-2]) + 4) + 'sp'
        width: dp(350)
        text: f'[b]{root.Name}[/b]'


<DSOPopup>:
    size_hint: None, 1
    width: dp(350)
    orientation: 'vertical'
    padding: dp(5), dp(2)


<Annotation>:
    size_hint: None, None
    size: self.texture_size
    font_size: '15sp' if self.pinned else '20sp'
    text_size: None, None  # forces size to be that of text
    color: (self.label_color if self.pinned else self.marker_color) + [self.visible]
    pos: (self.lab_x - self.width / 2, self.lab_y - self.height / 2) if self.center and not self.pinned else (self.lab_x, self.lab_y)


<AnnotCluster>:
    canvas:
        Color:
            rgb: 1, 1, 1
            a: 1 if self.visible else 0
        Line:
            circle: self.sx, self.sy, self.radius
            dash_offset: 16
            dash_length: 4
            width: 1


<AnnotQuasar>:
    canvas:
        Color:
            rgb: 1, 0, 0
            a: 1 if self.visible else 0
        Line:
            points: [self.sx - self.gap - self.length, self.sy, self.sx - self.gap, self.sy]
            width: 1
        Line:
            points: [self.sx, self.sy - self.gap - self.length, self.sx, self.sy - self.gap]
            width: 1


<AnnotFOV>:
    canvas:
        Color:
            rgb: .4, .4, .4
            a: 1 if self.visible else 0
        Line:
            points: [self.sx - self.length/2, self.sy, self.sx + self.length/2, self.sy]
            width: 1
    color: 1, .4, .4, 1

'''
)


class Annotation(Label, HoverBehavior):

    px = NumericProperty(0)  # image pixel centre
    py = NumericProperty(0)
    sx = NumericProperty(0)  # screen pixel centre
    sy = NumericProperty(0)

    ot = StringProperty('')
    ang = NumericProperty(45)
    visible = BooleanProperty(False)  # in eyepiece and within mag limit
    display = BooleanProperty(False)  # within mag limit
    pinned = BooleanProperty(False)
    marker_color = ListProperty([0.5, 0.3, 0.3])
    label_color = ListProperty([0.5, 0.5, 0.5])
    infoline = StringProperty('')
    info = DictProperty({})
   
    bx = NumericProperty(0)  #  pixel on boundary
    by = NumericProperty(0)
    lab_x = NumericProperty(0)
    lab_y = NumericProperty(0)
    center = BooleanProperty(True)
    mag = NumericProperty(0)
    radius = NumericProperty(0)

    # Members for drag behavior
    last_x = NumericProperty(0)
    last_y = NumericProperty(0)
    total_delta_x = NumericProperty(0)
    total_delta_y = NumericProperty(0)
    dragged = BooleanProperty(False)


    def on_enter(self, *args):
        if self.visible:
            self.display_DSOInfo()


    def on_leave(self, *args):
        if self.visible:
            self.close_DSOInfo()


    def on_touch_down(self, touch):
        if self.collide_point(*touch.pos):
            if self.visible:
                self.last_x, self.last_y = touch.pos
                return True
        return False


    def on_touch_move(self, touch):
        if self.visible and self.pinned and self.collide_point(*touch.pos):
            self.dragged = True
            x, y = touch.pos
            delta_x = x - self.last_x
            delta_y = y - self.last_y
            self.last_x, self.last_y = touch.pos
            self.x += delta_x
            self.y += delta_y
            self.lab_x += delta_x
            self.lab_y += delta_y

            # These are needed for when the view is dragged after dragging the label
            self.total_delta_x += delta_x
            self.total_delta_y += delta_y
            return True
        return False


    def on_touch_up(self, touch):
        if self.visible and self.collide_point(*touch.pos):
            if self.dragged:
                self.dragged = False
            else:
                if not self.pinned:
                    self.pin()
                else:
                    self.unpin()
                return True

        return False


    def pin(self, dt=None):
        if not self.pinned:
            lab = self.info['Name']
            self.font_size = '15sp'
            # for quasars, label with mag + redshift
            if self.ot == 'QS':
                lab = f'{self.info["Mag"]:4.1f}'
                z = self.info.get('z', '?')
                if not np.isnan(z):
                    lab += f' z {z:3.1f}'
                self.font_size = '14sp'
            # for members, use member number
            elif self.ot == 'ME':
                lab = self.info['Mem']
            self.text = str(lab)
            self.pinned = True
            self.close_DSOInfo()
            self.update()


    def unpin(self, dt=None):
        if self.pinned:
            if self.ot == 'QS':
                self.text = ' '
            else:
                self.text = '¤'
            self.pinned = False
            self.total_delta_x = 0
            self.total_delta_y = 0
            self.display_DSOInfo()
            self.update()


    def close_DSOInfo(self, *args):
        dsopopup.hide()


    def display_DSOInfo(self, *args):
        dsopopup.display(info=self.info, title_color=self.marker_color)


    def in_eyepiece(self, xc, yc, r2):
        return (xc - self.sx) ** 2 + (yc - self.sy) ** 2 < r2


    def update(self, mapping=None, xc=None, yc=None, r2=None):
        if xc is not None:
            self.xc = xc
        if yc is not None:
            self.yc = yc
        if mapping is not None:
            self.mapping = mapping
        if r2 is not None:
            self.r2 = r2
        self.sx, self.sy = self.mapping(self.px, self.py)
        self.visible = self.in_eyepiece(self.xc, self.yc, self.r2) and self.display

        # display if visible and set to display
        if self.visible:
            bx, by = self.mapping(self.bx, self.by)
            if self.pinned:
                self.radius = ((self.sx - bx) ** 2 + (self.sy - by) ** 2) ** 0.5
                ang_r = np.radians(self.ang)
                self.lab_x = float(self.sx + self.radius * np.cos(ang_r)) + self.total_delta_x
                self.lab_y = float(self.sy + self.radius * np.sin(ang_r)) + self.total_delta_y
            else:
                self.lab_x = self.sx
                self.lab_y = self.sy


class AnnotCluster(Annotation):
    pass


class AnnotQuasar(Annotation):
    # a couple of lines set a few pix from QSO
    length = NumericProperty(20)  # indicator line length in pixels
    gap = NumericProperty(20)  # gap from quasar to line in pixels


class AnnotFOV(Annotation):
    # represents FOV line
    length = NumericProperty(0)

    def update(self, mapping=None, **args):
        # compute length and update text based on FOV
        # sx is location
        if mapping is not None:
            self.mapping = mapping
        sx, sy = self.mapping(self.px, self.py)
        bx, by = self.mapping(self.bx, self.by)
        arcsecs = ((sx - bx) ** 2 + (sy - by) ** 2) ** 0.5
        self.length = arcsecs * 60
        self.sx = Window.width // 2
        self.sy = 200
        self.lab_x = self.sx
        self.lab_y = self.sy + 20
        self.visible = self.display


    def on_enter(self, *args):
        pass


    def on_touch_down(self, touch):
        return False


class TitleBar(BoxLayout):
    Name = StringProperty('')
    title_color = ListProperty([1, 0, 0, 1])


class PropVal(BoxLayout):
    param = StringProperty('')
    value = StringProperty('')


class DSOPopup(BoxLayout):

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        Window.add_widget(self)
        self.hide()


    def display(self, info=None, title_color=None):

        # ideally, exclude some of these in Catalogue
        exclude = {'OT', 'Name', 'RA', 'Dec', 'Con', 'Obs', 'List', 'Added', 'MaxAlt', 'Transit'}

        self.clear_widgets()
        tb = TitleBar(Name=info.get('Name', ''))
        if title_color is not None:
            tb.title_color = title_color
        self.add_widget(tb)

        for k, v in info.items():
            if str(v) and str(v) != 'nan' and k not in exclude:
                if type(v) == float:
                    pv = PropVal(param=k, value=f'{v:.3f}')
                else:
                    pv = PropVal(param=k, value=str(v))
                self.add_widget(pv)

        # position to popup info; not yet working in terms of limiting it to screen
        pos = Window.mouse_pos
        self.x = min(pos[0] + dp(9), Window.width - self.minimum_width - dp(20))
        self.y = min(pos[1], Window.height - self.minimum_height - dp(20))


    def hide(self):
        self.x = 10 * Window.width

# ensures that there is only one ie singleton popup
dsopopup = DSOPopup()


class Annotator(Component):

    mag_limit = NumericProperty(0)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.annotations = []


    def on_new_object(self):
        self.clear_annotations()
        self.on_mag_limit()


    def clear_annotations(self):
        # remove previous annotations
        gui = App.get_running_app().gui
        if self.has_annotations():
            n_annots = len(self.annotations)
            for k in self.annotations:
                gui.remove_widget(k)
            logger.info(f'Cleared {n_annots:d} annotations')
            self.annotations = []


    def has_annotations(self):
        return len(self.annotations) > 0


    def on_mag_limit(self, *args):

        # modify what is visible
        ml = self.mag_limit

        # if limit is below -5, turn off all annotations
        if ml <= -5:
            for annot in self.annotations:
                annot.display = False
            mesg = 'off'
        # just keep scale marker
        elif ml <= -3:
            for annot in self.annotations:
                annot.display = annot.info.get('Name', '') == 'FOV'
            mesg = 'fov'
        # if below 0, just turn on pinned annotations
        elif ml < 0:
            for annot in self.annotations:
                annot.display = annot.pinned
            mesg = 'pin'
        # otherwise apply magnitude limit to unpinned cases
        else:
            for annot in self.annotations:
                annot.display = annot.pinned or annot.mag < ml
            mesg = f'{ml:4.1f}'

        self.app.gui.set_prop('mag_limit', 'text', f'   {mesg}')

        self.update()


    @staticmethod
    def get_diam(info):
        if 'Diam' in info:
            d = info['Diam']
        elif 'diam' in info:
            d = info['diam']
        elif 'Major' in info:
            d = info['Major']
        else:
            d = .2
        if np.isnan(d):
            d = .2 
        return d


    def annotate(self, solution):
        ''' called by platesolver when finished a platesolve
        '''

        ''' solution is a dict like this:
            {
                'ra_centre': ra_centre,
                'dec_centre': dec_centre,
                'ra': str(RA(ra_centre)),
                'dec': str(Dec(dec_centre)),
                'width': w,
                'height': h,
                'fov_w': fov_w,
                'fov_h': fov_h,
                'north': north,
                'projection': projection,
                'im_stars': im_stars,
                'ref_stars': ref_stars
            }
        '''

        self.clear_annotations()
        # solver = Component.get('PlateSolver')
        cats = Component.get('Catalogues')

        # rotate view to north
        Component.get('View').orientation = solution['north']

        # find objects in FOV from tile large enough to encompass sensor
        w, h = solution['width'], solution['height']
        deg_per_pixel = float(solution['fov_h'] / h)
        fov = 1.5 * max(solution['fov_w'], solution['fov_h'])
        # possible that these should be tile ra/dec]
        ra0, dec0 = solution['tile_ra0'], solution['tile_dec0']
        tile = make_tile(ra0, dec0, fov=fov)

        dsos = cats.get_annotation_dsos(tile=tile)

        target = Component.get('DSO').Name

        # add FOV annotator
        self.annotations = [
            AnnotFOV(
                text="1'",
                info={'Name': 'FOV'},
                color=[0.4, 0.4, 0.4, 1],
                px=0,
                py=0,
                bx=1 / (3600 * deg_per_pixel),
                by=0,
                mag=-3,
                pinned=True
            )
        ]

        # object type properties and defaults
        OT_props = cats.get_object_types()
        default_props = {'name': '?', 'col': [0, 0, 0], 'text': '¤', 'mag': 0}

        ''' dsos is a dict of names and properties
        '''

        atypes = []
        for nm, info in dsos.items():
            px, py = radec2pix(info['RA'], info['Dec'], ra0, dec0, solution['projection'])

            # px, py = solver.ra_dec_to_pixels(info['RA'], info['Dec'])
            px, py = float(px[0]), float(py[0])
            # check it is actually within the image dimensions
            if (px >= 0) & (py >= 0) & (px < w) & (py < h):
                # replace OT acronym in info by full OT name
                ot = info['OT']
                atypes += [ot] # diagnostic
                props = OT_props.get(ot, default_props)
                # fill in missing props
                for k, v in default_props.items():
                    if k not in props:
                        props[k] = v
                info['Object type'] = props['name']
                # try to get a diameter value in case we need it
                diam = self.get_diam(info)
                ang = 45
                center = True
                if ot in {'OC', 'GC', 'CG', 'GG', 'G+', 'PG'}:
                    cls = AnnotCluster
                    diam *= 1.3  # make circle slightly larger
                    center = False
                elif ot in {'QS', 'BQ', 'DQ'}:
                    cls = AnnotQuasar
                else:
                    cls = Annotation
                    if ot in {'ME'}:
                        diam = .15   # put labels closer
                        ang = 0
                    elif ot in {'PN'}:
                        diam *= 1.5

                # diam is in arcmin
                pixel_rad = (((diam / (2 * 60.0)) / deg_per_pixel)) / 2**.5

                self.annotations += [cls(
                        ot=ot,
                        text=props['text'],
                        px=px, 
                        py=py,
                        info=info,
                        ang=ang,
                        center=center,
                        mag=info.get('Mag', props['mag']),
                        marker_color=props['col'],
                        bx=px + pixel_rad, 
                        by=py + pixel_rad)
                ]


        # create annotation labels
        mapping = Component.get('View').scatter.to_parent
        xc, yc = Metrics.get('origin')
        r2 = Metrics.get('inner_radius') ** 2
        gui = App.get_running_app().gui
        for lab in self.annotations:
            # gui.add_widget(lab, index=101)
            gui.add_widget(lab, index=0)
            lab.update(mapping=mapping, xc=xc, yc=yc, r2=r2)
            # sort out situations where several have same target name
            if lab.info['Name'] == target:
                lab.pin()
            ''' check if its a G2V and if so get image pixel values in RGB and report (for now)
            '''
            # if lab.ot == 'G2':
            #     px, py = lab.px, lab.py
            #     Component.get('Stacker').get_RGB_at_pixel(lab.px, lab.py)
                # print('found G2V im px {:} {:}'.format(lab.px, lab.py), props)

        self.on_mag_limit()


    def update(self):

        # called whenever a redraw is required e.g. View or maglimit changes
        if not self.has_annotations():
            return

        mapping = Component.get('View').scatter.to_parent
        xc, yc = Metrics.get('origin')
        r2 = Metrics.get('inner_radius') ** 2
        for annot in self.annotations:
            annot.update(mapping=mapping, xc=xc, yc=yc, r2=r2)




''' camera obeying ASCOM protocol (status: in development)
'''

import time
import numpy as np

from loguru import logger
from kivy.properties import StringProperty
from kivy.clock import Clock

from jocular.ascom import connect_to_ASCOM_device
from jocular.cameras.genericcamera import GenericCamera


class ASCOMCamera(GenericCamera):

	# configurables = {
	# 	'binning': {'options': ['None', '2 x 2', '3 x 3', '4 x 4'], 'help': 'some help text'},
	# 	'colour space': {'options': ['Mono', 'RGGB', 'GRGB', 'BGGR']},
	# 	'gain': {'range': (1, 1000), 'init': 150, 'int': True},
	# 	'cooling': {'range': (-40, 10), 'init': 0, 'int': True}
	# 	}

	driver = StringProperty(None)

	def on_close(self):
		self.disconnect()

	def disconnect(self):
		logger.debug('closing ASCOM camera')
		self.connected = False
		self.status = 'disconnected'
		self.camera.Connected = False  # closes ascom devic

	def connect(self):

		if self.connected:
			return

		res = connect_to_ASCOM_device(device_type='Camera', driver=self.driver)
		self.connected = res.get('connected', False)
		self.status = res['status']
		if self.connected:
			self.driver = res.get('driver', self.driver)
			self.camera = res['device']
			self.status += ': {:} x {:}'.format(
				self.camera.CameraXSize, self.camera.CameraYSize)

		else:
			if 'exception' in res:
				logger.exception('ascom connect problem ({:})'.format(res['exception']))


	def capture(self, exposure=None, on_capture=None, on_failure=None, is_faf=False,
		binning=None, return_image=False, is_bias=False):

		self.exposure = exposure
		self.on_failure = on_failure
		self.on_capture = on_capture

		if is_bias:
			self.exposure = .001
			openshutter = False
		else:
			openshutter = True

		self.camera.startExposure(self.exposure, openshutter) 

		if return_image:
			time.sleep(self.exposure)
			ready = False
			while not ready:
				time.sleep(.2)
				ready = self.camera.ImageReady
			return self.get_camera_data()

		self.capture_event = Clock.schedule_once(self.check_exposure, 
			max(0.2, self.exposure))

	def check_exposure(self, *arg):
		if self.camera.ImageReady:
			self.last_capture = self.get_camera_data()
			if self.on_capture is not None:
				Clock.schedule_once(self.on_capture, 0)
		else:
			self.capture_event = Clock.schedule_once(self.check_exposure, 0.2)

	def get_camera_data(self):
		# data is ready so get it, convert to correct size, and scale to range 0-1
		# assumes 16 bits for now
		im = np.array(self.camera.ImageArray).as_type(np.uint16) / 2**16
		return im.T

	def stop_capture(self):
		# Cancel any pending reads
		if hasattr(self, 'camera') and self.camera is not None:
			# occasionally objects to being aborted
			if self.camera.CanStopExposure:
				try:
					self.camera.StopExposure()
					logger.debug('stopped exposure')
				except Exception as e:
					logger.exception('unable to stop exposure ({:})'.format(e))
		if hasattr(self, 'capture_event'):
			self.capture_event.cancel()





''' Represents session info
'''

import json
import math
from datetime import datetime
from loguru import logger

from kivy.app import App
from kivy.properties import StringProperty, BooleanProperty, NumericProperty
from kivy.clock import Clock
from kivy.lang import Builder
from kivymd.uix.boxlayout import MDBoxLayout

from jocular.component import Component
from jocular.settingsmanager import JSettings
from jocular.widgets.widgets import Pin


Builder.load_string('''

<MyBoxLayout@BoxLayout>:
    size_hint: (1, None)
    height: '42dp'

<SessionInfo>:
    padding: '10dp'
    adaptive_height: True
    pos_hint: {'y': 0, 'x': 0} if root.session.show_session else {'y': 0, 'right': -1000} 
    size_hint: None, None
    orientation: 'vertical'
    spacing: "10dp"

    BoxLayout:
        size_hint: (1, None)
        height: '48dp'

        JTextField:
            width: '180dp'
            height: '32dp'
            on_focus: root.session.session_changed(self.text) if not self.focus else None
            text: root.session.session

    MyBoxLayout:
        JTextField:
            hint_text: 'transparency'
            width: '220dp'
            on_focus: root.session.transparency_changed(self.text) if not self.focus else None
            text: root.session.transparency


    MyBoxLayout:
        JTextField:
            hint_text: 'seeing'
            width: '240dp'
            on_focus: root.session.seeing_changed(self.text) if not self.focus else None
            text: root.session.seeing

    MyBoxLayout:
        JTextField:
            hint_text: 'temperature'
            on_focus: root.session.temperature_changed(self.text) if not self.focus else None
            text: root.session.formatted_temperature

        JTextField:
            width: '140dp'
            hint_text: 'brightness'
            on_focus: root.session.sky_brightness_changed(self.text) if not self.focus else None
            text: root.session.formatted_sky_brightness

    MyBoxLayout:
        JTextField:
            width: '140dp'
            hint_text: 'scope'
            #helper_text: ''
            on_focus: root.session.telescope_changed(self.text) if not self.focus else None
            text: root.session.telescope

        JTextField:
            width: '180dp'
            hint_text: 'camera'
            #helper_text: ''
            on_focus: root.session.camera_changed(self.text) if not self.focus else None
            text: root.session.camera

    BoxLayout:
        size_hint: (1, None)
        height: '72dp'
        JTextField:
            id: _name
            width: '400dp'
            height: '64dp'
            multiline: True
            hint_text: 'session notes'
            #helper_text: ''
            text: root.session.session_notes
            on_focus: root.session.session_notes_changed(self.text) if not self.focus else None

    BoxLayout:
        size_hint: (1, None)
        height: '22dp'

''')


date_time_format = '%d %b %y %H:%M'


def datenow():
    return datetime.now().strftime(date_time_format)


def hours_since_date(mydate):
    return (datetime.now() - datetime.strptime(mydate, date_time_format)).total_seconds()/3600


# from http://unihedron.com/projects/darksky/NELM2BCalc.html
def SQM_to_NELM(sqm):
    return 7.93 - 5 * math.log(10 ** (4.316-(sqm / 5)) + 1)


def NELM_to_SQM(nelm):
    return 21.58 - 5 * math.log(10 ** (1.586 - nelm / 5) - 1)



class SessionInfo(MDBoxLayout):
    def __init__(self, session, **kwargs):
        self.session = session
        super().__init__(**kwargs)



class Session(Component, JSettings):

    save_settings = ['show_session']

    session = StringProperty('')
    sky_brightness = NumericProperty(None, allownone=True)
    temperature = NumericProperty(None, allownone=True)
    seeing = StringProperty('')
    transparency = StringProperty('')
    session_notes = StringProperty('')
    telescope = StringProperty('')
    camera = StringProperty('')
    is_new_object = BooleanProperty(False)
    show_session = BooleanProperty(False)
    sky_brightness_units = StringProperty('SQM')
    temperature_units = StringProperty('Centigrade')
    formatted_temperature = StringProperty('')
    formatted_sky_brightness = StringProperty('')
    retain_hours =  NumericProperty(3)

    configurables = [
        ('retain_hours', {'name': 'time to retain session information?', 'float': (0, 48, 1),
            'help': 'Session information persists for some time to handle restarts of the program',
            'fmt': '{:.0f} hours'}),
        ('temperature_units', {
            'name': 'temperature units', 
            'options': ['Centigrade', 'Fahrenheit'],
            'help': 'temperatures are stored internally in degrees C'
            }),
        # ('sky_brightness_units', {
        #     'name': 'sky brightness units', 
        #     'options': ['SQM', 'NELM'],
        #     'help': 'brightness stored internally in SQM'
        #     })
        ]

    # all editable properties
    props = ['sky_brightness', 'temperature', 'seeing', 'transparency', 'session_notes', 
        'telescope', 'camera', 'session']


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        show_session = self.app.gui.get_gui_setting('show_session')
        self.show_session = False if show_session is None else show_session
        self.session_info = SessionInfo(self)
        self.app.gui.add_widget(self.session_info)
        self.app.gui.add_widget(Pin(
            comp=self, 
            field='show_session', 
            loc='lower-left', 
            tooltip_text='toggle session information panel',
            show_text='Session'))


    def describe(self):
        ''' return information for Snapshotter
        '''
        return [
            self.session, 
            self.formatted_sky_brightness,
            self.formatted_temperature,
            f'seeing {self.seeing}' if self.seeing else None,
            f'transp {self.transparency}' if self.transparency else None
            ]


    def theme_changed(self, *args):
        self.session_info.theme_changed()


    def temperature_changed(self, temp):
        ''' called when temperature field is altered; parse various formats
            with and without degrees and C or F
        '''
        temp = temp.lower().strip()
        try:
            if len(temp) == 0:
                self.temperature = None
            else:
                units = 'F' if temp.endswith('f') else 'C'
                if temp[-1] in {'f', 'c'}:
                    temp = temp[:-1].strip()
                if temp[-1] == '\N{DEGREE SIGN}':
                    temp = temp[:-1].strip()
                temp = int(temp)
                self.temperature = temp if units == 'C' else (temp - 32) / 1.8
        except Exception as e:
            self.temperature = None
            logger.exception(e)
        self.on_temperature()
        self.check_for_change() 


    def get_temperature(self):
        ''' return string temperature in C or F, or empty string if error
            in conversion or non-existent temperature
        '''
        try:
            temp = int(self.temperature)
        except:
            return ''
        if self.temperature_units == 'Centigrade':
            return f'{temp:.0f}\N{DEGREE SIGN}C'
        return f'{temp*1.8 + 32:.0f}\N{DEGREE SIGN}F'


    @logger.catch()
    def on_temperature(self, *args):
        ''' whenever temperature changes, ensure formatted temperature
            is updated
        '''
        self.formatted_temperature = 'xyz'  # forces an update
        self.formatted_temperature = self.get_temperature()


    def on_temperature_units(self, *args):
        ''' whenever user changes temperature preferences in config screen,
            ensure it is applied immediately
        '''
        self.formatted_temperature = self.get_temperature()


    ''' similar methods for sky brightness
    '''

    def sky_brightness_changed(self, bright):
        bright = bright.lower().strip()
        try:
            if len(bright) == 0:
                self.sky_brightness = None
            else:
                units= 'SQM'
                if bright.endswith('nelm'):
                    units = 'NELM'
                    bright = bright[:-4].strip()
                elif bright.endswith('sqm'):
                    bright = bright[:-3].strip()
                bright = float(bright)
                self.sky_brightness = bright if units == 'SQM' else SQM_to_NELM(bright)
        except:
            self.sky_brightness = None
        self.on_sky_brightness()
        self.check_for_change() 


    def get_sky_brightness(self):
        try:
            bright = float(self.sky_brightness)
        except:
            return ''
        if self.sky_brightness_units == 'SQM':
            return f'{bright:.2f} sqm'
        return f'{NELM_to_SQM(bright):.1f} nelm'


    def on_sky_brightness(self, *args):
        self.formatted_sky_brightness = 'xyz'
        self.formatted_sky_brightness = self.get_sky_brightness()


    def on_sky_brightness_units(self, *args):
        self.formatted_sky_brightness = self.get_sky_brightness()


    ''' remaining changes
    '''  

    def session_notes_changed(self, session_notes):
        self.session_notes = session_notes
        self.check_for_change() 


    def seeing_changed(self, seeing):
        self.seeing = seeing
        self.check_for_change() 


    def transparency_changed(self, transparency):
        self.transparency = transparency
        self.check_for_change() 


    def telescope_changed(self, telescope):
        ''' Since we persist this between sessions, ensure we save
            the session on any change; likewise with camera
        '''
        self.telescope = telescope
        self.check_for_change()
        self.save_session()


    def camera_changed(self, camera):
        self.camera = camera
        self.check_for_change()
        self.save_session() 


    def session_changed(self, session):
        self.session = session
        self.check_for_change() 


    @logger.catch()
    def check_for_change(self):
        ''' Check if any property has changed and stack is not empty
            For new objects, don't check if session has changed as clock is
            continually changing! 
        '''
        props = self.props
        if self.is_new_object:
            props = set(props) - {'session'}

        changes = [getattr(self, p) != self.initial_values.get(p, '') for p in props]
        self.changed = 'session props' if any(changes) else ''


    ''' main methods for handling new/previous sessions
    '''

    @logger.catch()
    def on_new_object(self):
        self.is_new_object = True
        self.cancel_session_clock()
        self.load_session()
        self.initial_values = {p: getattr(self, p) for p in self.props}


    @logger.catch()
    def on_previous_object(self):
        ''' apply previous session settings, keeping a record so changes can be monitored
        '''
        self.is_new_object = False
        self.cancel_session_clock()
        settings = Component.get('Metadata').get(self.props)
        self.apply_session_settings(settings)
        self.initial_values = {p: getattr(self, p) for p in self.props}


    def cancel_session_clock(self):
        if hasattr(self, 'session_clock'):
            self.session_clock.cancel()


    def load_session(self):
        ''' Try to get information from last session and apply it, otherwise
            create a new session
        '''
        try:
            with open(self.app.get_path('last_session.json'), 'r') as f:
                session = json.load(f)
            if hours_since_date(session['session']) < self.retain_hours:
                self.apply_session_settings(session)
            else:
                # start fresh session but restore scope/camera details
                self.create_empty_session()
                self.telescope = session.get('telescope', '') 
                self.camera = session.get('camera', '')
        except Exception as e:
            # any problems we just start a fresh session
            logger.debug(f'problem loading session: recreating {e}')
            self.create_empty_session()
            self.save_session()

        # create date and clock to update it
        self.session = datetime.now().strftime(date_time_format)
        self.session_clock = Clock.schedule_interval(self.update_time, 1.0)


    def create_empty_session(self):
        ''' set all properties except session to empty/None
        '''
        self.temperature = None
        self.sky_brightness = None
        for p in set(self.props) - {'session', 'temperature', 'sky_brightness'}:
            setattr(self, p, '')


    def get_session(self):
        # accessor method that strips off seconds if present
        if self.session.count(':') == 2:
            return self.session[:-3]
        else:
            return self.session


    def current_session_info(self):
        session = {}
        for p in self.props:
            v = getattr(self, p)
            if v:
                session[p] = v
        session['session'] = self.get_session()
        return session


    def apply_session_settings(self, settings):
        ''' apply 'settings', by clearing existing settings and
            overriding
        '''
        self.create_empty_session()
        for p in self.props:
            if p in settings:
                setattr(self, p, settings[p])

    @logger.catch()
    def save_session(self):
        logger.debug('saving session')
        with open(self.app.get_path('last_session.json'), 'w') as f:
            json.dump(self.current_session_info(), f, indent=1)    


    @logger.catch()
    def on_save_object(self):

        logger.debug('saving properties to metadata')

        # save properties
        for p in self.props:
            Component.get('Metadata').set(p, getattr(self, p))

        # only update last session if live rather than previous
        if self.is_new_object:
            self.save_session()

    def update_time(self, *args):
        self.session = datetime.now().strftime('%d %b %y %H:%M:%S')



''' Handles load/save of all settings. 
'''

import json
import os
from functools import partial
from loguru import logger

from kivy.app import App
from kivy.uix.label import Label
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.spinner import Spinner
from kivy.metrics import dp
from kivy.properties import StringProperty, DictProperty, ListProperty
from kivy.factory import Factory
from kivy.lang import Builder

from jocular.component import Component
from jocular.formwidgets import configurable_to_widget
from jocular.panel import Panel

Builder.load_string(
    '''

<MySpinnerOption@SpinnerOption>:
    size_hint_y: None
    height: "36dp"

''')

class SettingsManager(Component, Panel):


	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.app = App.get_running_app()
		self.instances = {}
		self.current_panel = 'Appearance'
		self.app.gui.add_widget(self)


	def register(self, settings, name=None):
		''' keep a record of all setting instances
		'''
		self.instances[name] = settings
		logger.debug(f'registered settings for {name}')


	def on_hide(self, *args):
		for cls in self.instances.values():
			cls.apply_and_save_settings()


	def on_show(self, panel=None):
		''' Rebuild and display settings screen
		'''

		self.contents.clear_widgets()
		self.header.clear_widgets()
		self.contents.width = dp(800) # was 600

		# top spinner
		hb = BoxLayout(size_hint=(1, None), height=dp(28))
		hb.add_widget(Label(size_hint=(1, 1)))
		self.spinner = Spinner(
			text=self.current_panel,
			values=sorted(self.instances.keys()),
			option_cls=Factory.get("MySpinnerOption"),
			size_hint=(None, 1), width=dp(140), 
			font_size='16sp')
		self.spinner.bind(text=self.setting_panel_changed)
		hb.add_widget(self.spinner)
		hb.add_widget(Label(size_hint=(1, 1)))
		self.header.add_widget(hb)

		self.panel = BoxLayout(orientation='vertical', size_hint=(1, 1))
		self.contents.add_widget(self.panel)
		self.show_panel()


	def show_settings(self, panel):
		if panel not in self.instances:
			return
		self.show()
		self.spinner.text = panel


	def setting_panel_changed(self, spinner, *args):
		self.current_panel = spinner.text
		self.show_panel()


	def show_panel(self):
		''' update panel with current class settings
		'''

		cls = self.instances[self.current_panel]

		self.panel.clear_widgets()
		self.panel.add_widget(Label(size_hint=(1, 1))) # spacer

		for name, spec in cls.configurables:
			self.panel.add_widget(configurable_to_widget(
				text=spec.get('name', name),
				name=name,
				spec=spec,
				helptext=spec.get('help', ''),
				initval=getattr(cls, name), 
				changed=cls.setting_changed))

		self.panel.add_widget(Label(size_hint=(1, 1))) # spacer


	def on_touch_down(self, touch):
		handled = super().on_touch_down(touch)
		if self.collide_point(*touch.pos):
			return True
		return handled



class SettingsBase():

	configurables = ListProperty([])
	changed_settings = DictProperty({})


	def __init__(self, **kwargs):
		self.app = App.get_running_app()
		self.name = self.__class__.__name__

		settings = {}

		# load settings if they exist
		path = self.app.get_path(f'{self.name}.json')
		if os.path.exists(path):
			# report any JSON issues
			try:
				with open(path, 'r') as f:
					settings = json.load(f)
			except Exception as e:
				logger.error(f'Problem reading json file {path} ({e})')
		else:
			logger.debug(f'No {path}; setting to empty')


		# initialise values; report any issues caused e.g. by changes 
		# in types etc between versions
		for name, spec in self.configurables:
			if name in settings:
				try:
					setattr(self, name, settings[name])
				except Exception as e:
					logger.debug(f'{e}')


	@logger.catch
	def apply_and_save_settings(self):

		if not self.changed_settings:
			return

		# update values
		for p, v in self.configurables:
			if 'action' not in v and p in self.changed_settings:
				setattr(self, p, self.changed_settings[p])

		# save settings
		settings = {p: getattr(self, p) for p, v in self.configurables if 'action' not in v}
		with open(self.app.get_path(f'{self.name}.json'), 'w') as f:
			json.dump(settings, f, indent=1)

		# notify anyone who is listening
		self.settings_have_changed() 

		# clear changes for next time
		self.changed_settings = {}


	@logger.catch()	
	def setting_changed(self, name, value, spec, *args):
		''' Called by widget when a setting changes; we just store it
		'''

		# either an action or a setting change
		if 'action' in spec:
			try:
				widget = spec['widget']
				widget.text = 'processing...'
				getattr(self, spec['action'])(
					success_callback=partial(self.action_succeeded, widget),
					failure_callback=partial(self.action_failed, widget)
					)
			except Exception as e:
				logger.exception(e)

		else:
			self.changed_settings[name] = value
			if spec.get('update', True):
				setattr(self, name, value) 


	def action_succeeded(self, widget, message, *args):
		widget.text = message


	def action_failed(self, widget, message, *args):
		widget.text = message


	def settings_have_changed(self, *args):
		pass



class JSettings(SettingsBase):

	tab_name = StringProperty(None) 

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		Component.get('SettingsManager').register(self,
			name=self.name if self.tab_name is None else self.tab_name)




''' Manages status panel showing information about a subset of components
'''

from kivy.app import App
from kivy.lang import Builder
from kivymd.uix.label import MDLabel
from kivymd.uix.boxlayout import MDBoxLayout
from kivy.properties import BooleanProperty

from jocular.component import Component
from jocular.widgets.widgets import Pin

Builder.load_string('''

<StatusLabel>:
    size_hint: 1, None
    height: dp(int(app.form_font_size[:-2]) + 2)
    halign: 'right'
    markup: True
    font_size: app.form_font_size
    text_size: self.size
    size: self.texture_size
    padding: dp(3), dp(1)
    shorten: True
    theme_text_color: 'Custom'
    text_color: app.theme_cls.accent_color

<Status>:
    size_hint: None, None
    orientation: 'vertical'
    padding: dp(5), dp(2)
    size: '300dp', app.gui.height / 2
    pos: app.gui.width - self.width if root.show_status else app.gui.width, 0
    Label:
        size_hint: 1, 1

<StatusBox>:
    component: _component
    size_hint: 1, None
    height: dp(40)
    orientation: 'vertical'
    padding: dp(10)
    spacing: dp(10)
    MDLabel:
        id: _component
        size_hint: 1, .3
        halign: 'right'
        font_size: '12sp'
        color: app.hint_color
''')


class StatusLabel(MDLabel):
   pass


class StatusBox(MDBoxLayout):
    pass


class Status(MDBoxLayout, Component):

    save_settings = ['show_status']

    show_status = BooleanProperty(False)
    comps = [
        'Capture', 'Calibrator', 'View', 'Aligner', 
        'PlateSolver', 'Monochrome', 'MultiSpectral', 'Stacker']


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        app = App.get_running_app()
        app.gui.add_widget(self, index=2)

        show_status = app.gui.get_gui_setting('show_status')
        self.show_status = False if show_status is None else show_status

        # add a status label for each component where we are interested in status updates
        self.labs = {c: StatusLabel(text='') for c in self.comps}
        for name, lab in self.labs.items():
            w = StatusBox()
            w.component.text = name.lower()
            w.add_widget(lab)
            self.add_widget(w)

        # add an empty widget
        self.add_widget(StatusBox())

        app.gui.add_widget(Pin(
            comp=self, 
            field='show_status', 
            loc='lower-right', 
            tooltip_text='toggle status panel',
            show_text='Status'))


    def bind_status(self, name, attr):
        # called by Component when a component is loaded
        if name in self.labs:
            attr.bind(infoline=self.labs[name].setter('text'))



''' Observing notes
'''

from kivy.app import App
from kivy.properties import StringProperty, BooleanProperty
from kivy.lang import Builder
from kivymd.uix.boxlayout import MDBoxLayout
from jocular.component import Component
from jocular.widgets.widgets import Pin


Builder.load_string('''
<Notes>:
    observing_notes: _notes
    adaptive_height: True
    pos_hint: {'top': .99, 'right': .99} if root.show_notes else {'top': .99, 'right': 0} 
    size_hint: None, None
    width: dp(200) 

    JTextField:
        id: _notes
        width: dp(200)
        multiline: True
        hint_text: 'Observing notes'
        helper_text: ''
        helper_text_mode: 'on_focus'
        on_focus: root.notes_changed() if not self.focus else None
        font_size: app.form_font_size

''')


class Notes(MDBoxLayout, Component):

    notes = StringProperty('')
    show_notes = BooleanProperty(False)
    save_settings = ['show_notes']


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        show_notes = self.app.gui.get_gui_setting('show_notes')
        self.show_notes = False if show_notes is None else show_notes
        self.app.gui.add_widget(self)
        self.app.gui.add_widget(Pin(
            comp=self, 
            field='show_notes', 
            loc='upper-right', 
            tooltip_text='toggle note-taking panel',
            show_text='Notes'))


    def on_new_object(self):
        self.notes = ''
        self.orig_notes = Component.get('Metadata').get('Notes', default='')
        self.observing_notes.text = self.orig_notes
        self.notes = self.observing_notes.text        


    def on_save_object(self):
        Component.get('Metadata').set('Notes', self.notes.strip())


    def notes_changed(self, *args):
        self.notes = self.observing_notes.text
        self.changed = '' if self.notes == self.orig_notes else 'observing notes'



''' Special-purpose widgets for Jocular
'''


from kivy.core.window import Window
from kivy.lang import Builder
from kivy.uix.label import Label
from kivy.properties import (
    NumericProperty,
    BooleanProperty
)

from jocular.uranography import angle_diff
from jocular.widgets.widgets import JRotWidget


Builder.load_string('''
<JLever>:
    color: app.lever_color
    font_size: '17sp'
''')


class JLever(JRotWidget, Label):
    ''' circular sliders
    '''

    value = NumericProperty(0.0)
    disabled = BooleanProperty(False)


    def __init__(self, value=0, values=None, angles=None, radial=True,
        continuous_update=None, shortcut=None, sc_inc=0.0, **kwargs):

        super().__init__(radial=radial, **kwargs)
        if values is None:
            values = [0, 1]
        if angles is None:
            angles = [0, 1]
        self.min_angle, self.max_angle = angles[0], angles[1]
        self.min_value, self.max_value = values[0], values[1]
        self.value = value
        self.selected = False

        # disabled this because can't get View on init reliably it seems
        # if continuous_update is None:
        #     self.update_on_move = Component.get('View').continuous_update # True
        # else:
        #     self.update_on_move = continuous_update
        self.update_on_move = True

        self._k = (self.max_angle - self.min_angle) / (self.max_value - self.min_value)
        self.rotangle = self.value_to_angle(self.value)

        # Use a simple helper class for keyboard shortcuts.
        # JLever has already been bound in a superclass to the touch handlers,
        # and cannot be itself rebound
        self.keyhandler = JLeverKeyHandler(self, shortcut)

        # Factor to increment by when shortcut is used
        self.sc_inc = sc_inc

        self.update(self.rotangle)


    def reset_value(self, v):
        # force update of lever position
        ang = self._k * (v - self.min_value) + self.min_angle
        self.update(ang)


    def value_to_angle(self, v):
        return self._k * (v - self.min_value) + self.min_angle


    def on_touch_down(self, touch):
        if not self.disabled and self.collide_point(*touch.pos):
            if touch.is_double_tap:
                self.update_on_move = not self.update_on_move
            self.selected = True
            return True
        return False


    def on_touch_move(self, touch):
        ''' If this lever is selected, update position, and update value if
            update_on_move is True
        '''
        if self.selected:
            self.update(
                angle_diff(self.origin, touch.pos),
                update_value=self.update_on_move)


    def ang2value(self, angle):
        # x = float(((angle - self.min_angle) / self._k) + self.min_value)
        x = ((angle - self.min_angle) / self._k) + self.min_value
        return min(self.max_value, max(x, self.min_value))


    # Called by the keyboard shortcut handler
    def increment(self, sign):
        angle = self.angle + sign * self.sc_inc
        self.update(angle, update_value=True)


    def on_touch_up(self, touch):
        ''' Unselect lever and update value if update_on_move is False
        '''
        if self.selected:
            self.selected = False
            if not self.update_on_move:
                _ang = angle_diff(self.origin, touch.pos)
                self.value = self.ang2value(_ang)
                # fake a transparency change to ensure interface is updated
                _t = self.color[-1]
                self.color[-1] = 0
                self.color[-1] = _t


    def update(self, angle, update_value=True):
        ''' If angle within range, update angle of level, and
            update corresponding value if update_value is True
        '''

        # return if angle outside range
        if (angle >= min(self.min_angle, self.max_angle)) and \
            (angle <= max(self.min_angle, self.max_angle)):
            if update_value:
                self.value = self.ang2value(angle)
                # convert value to angle in case it hasn't changed
                angle = self.value_to_angle(self.value)
            self.angle = angle
            self.on_angle()



# Helper class to bind to keyboard events
class JLeverKeyHandler:

    def __init__(self, jlever, shortcut):
        self.jlever = jlever
        self.shortcut = shortcut
        Window.bind(on_key_down=self.on_key_down)

    # Window.on_key_down has an undocumented fifth argument. A test program
    # shows this is the correct set of arguments.
    def on_key_down(self, window, key, scancode, codepoint, modifiers):
        if not (self.jlever.disabled or self.shortcut == None):
            if codepoint == self.shortcut and 'ctrl' in modifiers:
                self.jlever.increment(1)
            elif codepoint == self.shortcut and 'alt' in modifiers:
                self.jlever.increment(-1)



''' Manages previous observations and associated observations table.
'''

import os
import shutil
import glob
import json
from loguru import logger

from kivy.app import App
from kivy.clock import Clock
from kivy.core.window import Window

from jocular.component import Component
from jocular.table import Table
from jocular.metadata import get_metadata
from jocular.utils import toast
from jocular.image import Image, fits_in_dir
from jocular.exposurechooser import exp_to_str


class Observations(Component):


    def __init__(self):
        super().__init__()
        self.app = App.get_running_app()
        Clock.schedule_once(self.load_observations, 0)


    def get_observations(self):
        if not hasattr(self, 'observations'):
            self.load_observations()
        return self.observations


    def load_observations(self, dt=None):
        ''' Load json file containing observation details. If any problem,
            set observations to empty and continue.
        '''
        obspath = self.app.get_path('previous_observations.json')
        if os.path.exists(obspath):
            try:
                with open(obspath, 'r') as f:
                    self.observations = json.load(f)
                logger.info(f'Loaded {len(self.observations)} observations')
            except Exception as e:
                logger.warning(f'none loaded, rescanning captures ({e})')
                self.observations = {}
        # no obs so rebuild from captures
        # could be slow...
        else:
            self.get_observations_from_captures(detailed=False)


    def save_observations(self):
        try:
            with open(self.app.get_path('previous_observations.json'), 'w') as f:
                json.dump(self.observations, f, indent=1)
            logger.info(f'saved {len(self.observations)}')
        except Exception as e:
            logger.warning(f'cannot save previous observations ({e})')


    def get_observations_from_captures(self, detailed=True):
        ''' look in captures and its subdirectories to build observations structure
        '''
        self.observations = {}
        for sesh in glob.glob(os.path.join(self.app.get_path('captures'), '*')):
            for dso in glob.glob(os.path.join(sesh, '*')):
                if os.path.isdir(dso):
                    self.observations[dso] = get_metadata(dso, simple=True)
                    if detailed:
                        fail = 0
                        filts, expos, shapes = [], [], []
                        for f in fits_in_dir(dso):
                            try:
                                im = Image(f)
                                filts += [im.filter]
                                expos += [im.exposure]
                                shapes += [im.shape_str]
                            except:
                                fail += 1
                        if fail > 0:
                            logger.debug(f'{fail} unreadable FITs for {dso}')

                        # filters
                        filts = set(filts)
                        if {'L', 'R', 'G', 'B'} <= filts:
                            fstr = 'LRGB'
                        elif 'L' in filts:
                            rest = filts - {'L'}
                            fstr = 'L' + ''.join(rest)
                        else:
                            fstr = ''.join(filts)
                        self.observations[dso]['Filts'] = fstr

                        # exposures
                        expos = list(set(expos))
                        if len(expos) == 0:
                            expostr = ''
                        elif len(expos) > 1:
                            expostr = 'mult'
                        elif expos[0] is None:
                            expostr = ''
                        else:
                            expostr = exp_to_str(expos[0])
                        self.observations[dso]['Expo'] = expostr

                        shapes = list(set(shapes))
                        if len(shapes) != 1:
                            logger.debug(f'multiple sub dims for {dso}')
                            shapestr = 'mult'
                        else:
                            shapestr = shapes[0]
                        self.observations[dso]['Size'] = shapestr

        self.save_observations()


    def rebuild_observations(self, *args):
        ''' rebuild observations from scratch
        '''
        try:
            self.get_observations_from_captures(detailed=True)
            Clock.schedule_once(self.rebuild_table, .1)
        except Exception as e:
            logger.exception(e)
           

    def fast_rebuild_observations(self, *args):
        ''' rebuild observations from scratch
        '''
        try:
            self.get_observations_from_captures(detailed=False)
            Clock.schedule_once(self.rebuild_table, .1)
        except Exception as e:
            logger.exception(e)


    def rebuild_table(self, dt=None):
        self.observations_table.data = self.get_observations()         
        self.observations_table.update()
        toast('rebuilt!')


    def build_observations(self):
        # table construction ~150ms

        cols = {
            'Name': {'w': 250, 'align': 'left', 'sort': {'catalog':''}, 'action': self.load_dso},
            'OT': {'w': 40},
            'Con': {'w': 50},
            'Session': {'w': 140, 'sort': {'DateFormat': '%d %b %y %H:%M'}, 'label': 'Date'},
            'N': {'w': 40, 'align': 'right', 'type': int},
            'Expo': {'w': 70, 'sort': {'exposure': ''}},
            'Filts': {'w': 60},
            'Size': {'w': 100},
            'Notes': {'w': 1, 'align': 'left'}
        }

        # set up editable fields
        for c in ['Expo', 'Filts']:
            cols[c]['action'] = self.edit_prop

        return Table(
            size=Window.size,
            data=self.get_observations(),
            name='Observations',
            description='click on DSO name to load',
            cols=cols,
            actions={
                'move to delete dir': self.move_to_delete_folder,
                'rebuild observations': self.rebuild_observations,
                'fast rebuild': self.fast_rebuild_observations},
            on_hide_method=self.app.table_hiding,
            initial_sort_column='Session', 
            reverse_initial_sort=True
            )    


    def edit_prop(self, row, prop, value):
        # placeholder for now
        toast(f'editing prop {prop} {value} (not yet implemented)')


    def show_observations(self, *args):

        if not hasattr(self, 'observations_table'):
            self.observations_table = self.build_observations()
        self.app.showing = 'observations'

        # redraw on demand as it is expensive
        if self.observations_table not in self.app.gui.children:
            self.app.gui.add_widget(self.observations_table, index=0)

        self.observations_table.show()    


    def on_close(self, *args):
        ''' Save observations, clean up any empty observation and session dirs
        '''

        self.save_observations()

        capdir = self.app.get_path('captures')
        for sesh in glob.glob(os.path.join(capdir, '*')):
            dsos = os.listdir(sesh)
            if len(dsos) == 0 or ('.DS_Store' in dsos and len(dsos) == 1):
                try:
                    shutil.rmtree(sesh)
                    logger.debug(f'removing empty session {sesh}')
                except Exception as e:
                    logger.error(f'problem removing empty session {sesh} ({e})')
            else:
                for dso in dsos:
                    dname = os.path.join(sesh, dso)
                    if os.path.isdir(dname):
                        subs = os.listdir(dname)
                        if len(subs) == 0 or ('.DS_Store' in subs and len(subs) == 1):
                            try:
                                shutil.rmtree(dname)
                                logger.debug(f'removing empty DSO dir {dname}')
                            except Exception as e:
                                logger.error(f'problem removing empty DSO dir {dname} ({e})')


    def move_to_delete_folder(self, *args):
        objio = Component.get('ObjectIO')
        for s in self.observations_table.selected:
            if s in self.observations:
                objio.delete_file(s)
                del self.observations[s]
        logger.info(f'Deleted {len(self.observations_table.selected)} observations')
        self.observations_table.update()
        self.save_observations()


    def update(self, oldpath, newpath):
        ''' We have either an existing observation being saved or a 
            new observation, so read it and generate extra required info to 
            update self.observations
        '''

        # if not yet built, no worries because it will be found when next built
        if not hasattr(self, 'observations'):
            return

        self.observations[newpath] = get_metadata(newpath, simple=True)
        logger.debug(f'new path {newpath} = {self.observations[newpath]}')
        # user has changed directory
        if oldpath != newpath:
            if oldpath in self.observations:
                del self.observations[oldpath]

        # only update if we have already displayed table
        if hasattr(self, 'observations_table'):
            self.observations_table.update()

        self.save_observations()


    def load_dso(self, row, col, value):
        self.observations_table.hide()
        # convert row.key to str (it is numpy.str by default)
        Component.get('ObjectIO').load_previous(path=str(row.key))



''' Runs the capture scripts.
'''

import time
import numpy as np
from functools import partial
from scipy.interpolate import interp1d
from loguru import logger

from kivy.app import App
from kivy.properties import NumericProperty, BooleanProperty
from kivy.clock import Clock

from jocular.component import Component
from jocular.utils import toast, percentile_clip
from jocular.processing.starextraction import extract_stars

capture_controls = {'devices', 'script_button', 'capturing', 'exposure_button', 'filter_button'}


class Capture(Component):

    capturing = BooleanProperty(False)
    exposure = NumericProperty(0)


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.gui = self.app.gui


    def on_new_object(self):
        self.reset()
        self.info('inactive')
        self.gui.enable(['capturing'])


    def on_previous_object(self):
        self.info('inactive')
        self.gui.disable(['capturing'])


    def reset(self, stop_capturing=True):
        logger.debug(f'stop capturing? {stop_capturing}')
        self.series_number = None
        self.fps = 0
        self.last_faf = None
        if stop_capturing:
            self.stop_capture()


    def on_capturing(self, *args):
        # user (pressing camera button) or system changes capture state

        logger.debug('capturing state changed')

        if not self.capturing:
            self.stop_capture()
            return

        if not Component.get('Camera').connected():
            self.stop_capture(message='cannot connect to camera')
            return
            
        if not Component.get('FilterWheel').connected():
            self.stop_capture(message='cannot connect to filterwheel')
            return
            
        self.gui.disable(capture_controls)
        self.gui.disable({'load_previous', 'new_DSO'})
        self.gui.disable({'apply_ROI'})
        self.gui.enable({'capturing'})
        logger.debug('camera connected, starting capture')

        try:
            self.capture()
        except Exception as e:
            self.stop_capture(message=e)


    def stop_capture(self, message=None):
        # stops capture normally or abnormally
        logger.debug('stopping capture')
        Component.get('Camera').stop_capture()
        self.gui.set('capturing', False, update_property=True)
        self.gui.enable(capture_controls)
        self.gui.enable({'new_DSO', 'apply_ROI'})
        self.gui.enable({'load_previous'})
        if message is not None:
            Component.get('CaptureScript').reset_generator()
            logger.error(f'problem capturing ({message})')
            toast(f'Capture problem: {message}')
        self.info('stopped')


    def capture(self, *args):
        # generator yields next command to execute

        # check if user has pressed pause/stop
        if not self.capturing:
            return 

        # get next command from generator
        op = next(Component.get('CaptureScript').generator)

        # logger.debug(op)

        if len(op) == 2:
            op, param = op

        # allow ROI selection during framing captures
        # not yet working on ASI
        # if op == 'expose short':
        #     self.gui.enable({'apply_ROI'})
        # else:
        #     self.gui.disable({'apply_ROI'})

        # change filter
        if op == 'set filter':
            Component.get('FilterWheel').select_filter(name=param, 
                changed_action=self.capture,
                not_changed_action=partial(self.stop_capture, message='FW problem'))

        elif op == 'set exposure':
            self.exposure = param
            self.capture()

        # carry out a normal exposure
        elif op == 'expose long':
            try:
                self.start_capture_time = time.time()
                Component.get('Camera').capture_sub(
                    exposure=self.exposure, 
                    on_capture=self.save_capture,
                    on_failure=self.stop_capture)
                if self.exposure > 2:
                    self.expo = self.exposure
                    self.exposure_start_time = time.time()
                    Clock.schedule_once(self.tick, 0)
                self.info(f'capturing [{self.fps:.1f} fps]')
            except Exception as e:
                logger.exception(f'problem in expose long {e}')

        elif op == 'expose short':
            try:
                self.start_capture_time = time.time()
                Component.get('Camera').capture_sub(
                    exposure=self.exposure, 
                    on_capture=self.send_to_display,
                    on_failure=self.stop_capture,
                    is_faf=True)
                # for focus, info is updated in send_to_display
                if Component.get('CaptureScript').current_script != 'focus':
                    self.info(f'[{self.fps:.1f} fps]')  
            except Exception as e:
                logger.exception(f'problem in expose short {e}')

        elif op == 'expose bias':
            self.start_capture_time = time.time()
            Component.get('Camera').capture_sub(
                on_capture=self.save_capture,
                on_failure=self.stop_capture,
                is_bias=True)
            self.info(f'capturing bias [{self.fps:.1f} fps]')

        elif op == 'autoflat':
            auto_expo = self.get_flat_exposure()
            if auto_expo is not None:
                self.exposure = auto_expo
                self.capture()
            else:
                self.stop_capture()
                Component.get('CaptureScript').reset_generator()


    def on_exposure(self, *args):
        Component.get('CaptureScript').exposure_changed(self.exposure)        


    def tick(self, *args):
        ''' count up to exposure time using status info line
        '''
        dur = time.time() - self.exposure_start_time
        if self.capturing:
            self.info(f'Exposing {dur:2.0f}s [{self.fps:.1f} fps]')
            Clock.schedule_once(self.tick, 1)


    def send_to_display(self, *args):
        # send short subs directly to display
        self.fps = 1 / (time.time() - self.start_capture_time)
        try:
            im = Component.get('Camera').get_image()
            # new: store last short sub in case we want to platesolve
            self.last_faf = im
            fwhm = None
            if Component.get('CaptureScript').current_script == 'focus':
                # compute and show FWHM if possible, else clear
                try:
                    stars = extract_stars(im, star_method='photutils', target_stars=5)
                    fwhm = np.median(stars['fwhm'])
                    self.info(f'FWHM {fwhm:.1f}"/pix')
                except:
                    self.info('')
            Component.get('Monochrome').display_sub(im, fwhm=fwhm)
            self.capture()
        except Exception as e:
            logger.exception(f'{e}')


    def save_capture(self, *args):
        ''' Called via camera when image is ready.
        '''

        self.fps = 1 / (time.time() - self.start_capture_time)

        im = Component.get('Camera').get_image()
        capture_props = Component.get('Camera').get_capture_props()

        if im is None:
            toast('No image to save')
            return

        # image always comes across scaled 0-1 so rescale to 16-bits
        im *= (2**16 - 1)

        if not hasattr(self, 'series_number') or self.series_number is None:
            self.series_number = 1
        else: 
            self.series_number += 1

        capture_props['exposure'] = self.exposure
        filt = Component.get('FilterWheel').current_filter
        # map 'C' filter to 'L' at the point of capture
        capture_props['filter'] = 'L' if filt == 'C' else filt
        if capture_props.get('temperature', None) is None:
            capture_props['temperature'] = Component.get('Session').temperature
        sub_type = Component.get('CaptureScript').get_sub_type()
        capture_props['sub_type'] = sub_type

        # check why no bias in below (just a naming convention, but still...)
        prefix = sub_type if sub_type in {'flat', 'dark'} else capture_props['filter']
        name = f'{prefix}_{self.series_number:d}.fit'
        
        Component.get('ObjectIO').new_sub(
            data=im.astype(np.uint16),
            name=name,
            capture_props=capture_props)

        # ask for next capture immediately
        self.capture()


    def compute_ADU(self, expo):
        ''' Estimate ADU in central part of image given exposure
        '''
        im = Component.get('Camera').capture_sub(exposure=expo, 
            return_image=True,
            on_failure=self.stop_capture)
        return percentile_clip(im.ravel(), perc=75)


    def get_flat_exposure(self):
        ''' make a series of test exposures to get ADU (actually
            normalised to range 0-1) in the middle of the range; note
            that we only test exposures in the range 0.5-2.5s and aim
            for ADU in the range 0.3-0.7
        '''

        logger.info('doing autoflat')

        min_exposure, max_exposure = .5, 2.5
        min_ADU, max_ADU = .3, .7

        expos = np.linspace(min_exposure, max_exposure, 5)
        adus = np.ones(len(expos)) 

        # do shortest first in case no point
        adus[0] = self.compute_ADU(expos[0])
        if adus[0] > max_ADU:
            toast('Too early to collect flats')
            return None

        # do longest in case too late
        adus[-1] = self.compute_ADU(expos[-1])
        if adus[-1] < min_ADU:
            toast('Too late to collect flats')
            return None

        # do remaining ADUs
        adus[1: -1] = [self.compute_ADU(e) for e in expos[1: -1]]

        # interpolate to get things purrrrfect
        f = interp1d(expos, adus)
        adu_target = .7
        xvals = np.linspace(min_exposure, max_exposure, 500)
        best = np.argmin(np.abs(adu_target - f(xvals)))
        best_exposure = xvals[best]
        mesg = f'best exposure for autoflat {best_exposure} with ADU {f(best_exposure)}'
        toast(mesg)
        logger.info(mesg)
        return float(best_exposure)



''' Component to choose and apply stretch functions
'''

import numpy as np
from functools import partial

from skimage.exposure import (
    equalize_hist, equalize_adapthist, adjust_gamma, 
    adjust_log, adjust_sigmoid
    )
from skimage.filters import rank
from skimage.morphology import disk

from kivy.app import App
from kivy.metrics import dp
from kivy.uix.gridlayout import GridLayout
from kivy.uix.anchorlayout import AnchorLayout
from kivy.properties import StringProperty

from jocular.component import Component
from jocular.widgets.widgets import JMDToggleButton
from jocular.panel import Panel


class Stretcher(Panel, Component):

    stretch = StringProperty('asinh')


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.stretches = [
            'sublin', 'linear', 'gamma', 'log', 'asinh', 'hyper']
        # self.stretches = [
        #     'sublin', 'linear', 'hyper', 'log', 'gamma', 'asinh', 
        #     'hyper', 'histeq', 'clahe', 'rank',
        #     'gamma2', 'log2', 'ilog', 'sigmoid']
        self.build()
        self.panel_opacity = 0


    def on_show(self):
        for c, but in self.buttons.items():
            but.state = 'down' if c == self.stretch else 'normal'


    def _button(self, name):
        return JMDToggleButton(
                text=name, 
                size_hint=(1, None),
                group='stretch',
                font_size='16sp',
                on_press=partial(self.choose, name))


    def build(self, *args):

        content = self.contents
        self.buttons = {c: self._button(c) for c in self.stretches}
        layout = AnchorLayout(anchor_x='center', anchor_y='bottom', size_hint=(1, 1))
        content.add_widget(layout)
        gl = GridLayout(size_hint=(1, None), cols=6, height=dp(100), spacing=(dp(5), dp(5)))
        for b in self.buttons.values():
            gl.add_widget(b)
        layout.add_widget(gl)
        self.app.gui.add_widget(self)


    def choose(self, stretch, *args):
        if stretch != self.stretch:
            self.app.gui.set('stretch', stretch)
            self.stretch = stretch
            Component.get('Monochrome').adjust_lum()


    def apply_stretch(self, x, param=None, NR=1, background=None):
        ''' apply stretch, called in monochrome
        '''

        # if no noise reduction just use stretch alone
        if (NR <= 0) or (background is None):
            return stretch_main(x, method=self.stretch, param=param)

        # get stretched data and lightly suppress low end
        y = stretch_main(x, method=self.stretch, param=param)
        # y = y / np.max(y)
        hyper_param = 1 - .1 * (NR / 100)
        return y * stretch_main(x, method='hyper', param=hyper_param)


def stretch_main(x, method='linear', param=None):
    ''' Many of these are experimental and not exposed to the user
    '''

    if method == 'linear':
        return x

    if method == 'sigmoid':
        return adjust_sigmoid(x, cutoff=param)

    if method == 'gamma2':
        c = param
        return adjust_gamma(x, gamma=c)

    if method == 'log2':
        c = param
        return adjust_log(x, gain=c)

    if method == 'ilog':
        c = param
        return adjust_log(x, gain=c, inv=True)

    if method == 'histeq':
        return equalize_hist(x)

    if method == 'rank':
        footprint = disk(int((param + 2)*20))
        return rank.equalize(x, footprint=footprint)

    if method == 'clahe':
        return equalize_adapthist(x, clip_limit=param)

    if method == 'sublin':
        c = (param + 1)
        return x ** c

    if method == 'hyper':
        d = .02
        c = d * (1 + d - param)
        return (1 + c) * (x / (x + c))

    if method == 'log':
        c = param * 200
        return np.log(c*x + 1) / np.log(c + 1)

    if method == 'asinh':
        # c = param * 250
        c = param * 2000
        return np.arcsinh(c*x) / np.arcsinh(c + .0000001)

    if method == 'gamma':
        # with noise reduction, linear from x=0-a, with slope s
        y = x.copy()
        # g = .5 - .5 * param
        # g = .75 - .75 * param
        g = max(.01, 1 - param)
        a0 = .01
        s = g / (a0 * (g - 1) + a0 ** (1 - g))
        d = (1 / (a0 ** g * (g - 1) + 1)) - 1
        y[x < a0] = x[x < a0] * s
        y[x >= a0] =  (1 + d) * (x[x >= a0] ** g) - d
        return y
    
    return x




''' Saves currently displayed image in various formats.
'''

import mss
import os
from functools import partial
from datetime import datetime
import numpy as np
from skimage.io import imsave
from PIL import Image, ImageDraw, ImageFont
from loguru import logger

from kivy.app import App
from kivy.core.window import Window
from kivy.properties import StringProperty, BooleanProperty, NumericProperty
from kivy.metrics import Metrics as KivyMetrics
from kivy.metrics import dp
from kivy.uix.label import Label
from kivy.uix.button import Button
from kivy.uix.gridlayout import GridLayout
from kivy.clock import Clock

from jocular.component import Component
from jocular.settingsmanager import JSettings
from jocular.image import save_image
from jocular.metrics import Metrics
from jocular.utils import s_to_minsec, purify_name, toast
from jocular.widgets.widgets import JMDToggleButton
from jocular.widgets.icons import get_icon
from jocular.panel import Panel


class Snapshotter(Panel, Component, JSettings):

    save_settings= ['style', 'annotation', 'save_format']
    
    style = StringProperty('eyepiece')
    annotation = StringProperty('full')
    save_format = StringProperty('png')

    styles = {
        'eyepiece': 'circular image exactly as seen, with DSO labels if any',
        'landscape': 'rectangular image with no rotation/zoom/DSO labels'
    }

    annotations = {
        'plain': 'just the image',
        'name': 'image plus DSO name',
        'full': 'image plus all details (exposure, coords, date etc)'
    }

    save_formats = {
        'png': 'save as a .png format file',
        'jpg': 'save as a .jpg format file',
        'fits': 'save as a 16-bit FITs format file',
        'animated gif': 'save whole sequence of images as an animated gif'
    }

    imreduction = NumericProperty(1.8)
    framewidth = NumericProperty(0)
    dark_surround = BooleanProperty(True)

    configurables = [
        ('dark_surround', {
            'name': 'surround is', 
            'boolean': {'dark': True, 'light': False},
            'help': 'Shade of the region surrounding the eyepiece view'}),
        ('imreduction', {
            'name': 'image size reduction (eyepiece view)', 'float': (.5, 4, .1),
            'help': 'Reduces image size by this factor (useful on high-density screens)',
            'fmt': '{:.1f} x smaller'}),
        ('framewidth', {
            'name': 'add frame size around image', 'float': (0, 15, 1),
            'help': 'Adds a frame margin (landscape image only)',
            'fmt': '{:.0f} pixels'})
        ]


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.build()
        config = self.app.gui.config
        self.style = config.get('style', 'eyepiece')
        self.annotation = config.get('annotation', 'full')
        self.save_format = config.get('save_format', 'jpg')


    def on_show(self):
        for c, but in self.style_buttons.items():
            but.state = 'down' if c == self.style else 'normal'
        for c, but in self.annotation_buttons.items():
            but.state = 'down' if c == self.annotation else 'normal'
        for c, but in self.save_format_buttons.items():
            but.state = 'down' if c == self.save_format else 'normal'


    def _button(self, name, group, fn, tooltip=None):
        if tooltip is None:
            tooltip = group
        return JMDToggleButton(
                text=name, 
                size_hint=(1, None),
                group=group,
                font_size='16sp',
                tooltip_text=tooltip,
                on_press=partial(fn, name))


    def build(self, *args):

        content = self.contents
        content.add_widget(Label(text='', size_hint=(1, 1)))

        gl = GridLayout(size_hint=(1, None), cols=5, height=dp(40), spacing=(dp(5), dp(5)))
        self.style_buttons = {c: self._button(c, 'style', self.choose_style, tooltip=tt)
            for c, tt in self.styles.items()}
        for b in self.style_buttons.values():
            gl.add_widget(b)
        content.add_widget(gl)

        gl = GridLayout(size_hint=(1, None), cols=5, height=dp(40), spacing=(dp(5), dp(5)))
        self.annotation_buttons = {c: self._button(c, 'annotation', self.choose_annotation, tooltip=tt)
            for c, tt in self.annotations.items()}
        for b in self.annotation_buttons.values():
            gl.add_widget(b)
        content.add_widget(gl)

        gl = GridLayout(size_hint=(1, None), cols=5, height=dp(40), spacing=(dp(5), dp(5)))
        self.save_format_buttons = {c: self._button(c, 'save_format', self.choose_save_format, tooltip=tt)
            for c, tt in self.save_formats.items()}
        for b in self.save_format_buttons.values():
            gl.add_widget(b)
        content.add_widget(gl)

        content.add_widget(Label(text='', size_hint=(1, None), height=dp(30)))

        snap_text = get_icon('snapshotter', font_size=40) + '\n' + 'snap!'  
        snap_button = Button(
            text=snap_text,
            markup=True,
            background_color=(0, 0, 0, 0),
            size_hint=(1, None),
            height=dp(80),
            on_press=self.snap_no_return)

        content.add_widget(snap_button)

        self.app.gui.add_widget(self)


    def choose_style(self, style, *args):
        if style != self.style:
            self.style = style


    def choose_save_format(self, save_format, *args):
        if save_format != self.save_format:
            self.save_format = save_format


    def choose_annotation(self, annotation, *args):
        if annotation != self.annotation:
            self.annotation = annotation


    def mss_grabber(self, bbox=None):
        with mss.mss() as sct:
            im = sct.grab(bbox)
            return Image.frombytes('RGB', im.size, im.bgra, 'raw', 'BGRX').convert('RGBA')


    def large_font(self, w):
        f = max(18, int(w / 32))
        font = ImageFont.truetype(self.app.get_path('Roboto-Medium.ttf'), f)
        (width, height), (offset_x, offset_y) = font.font.getsize('A')
        return font, f, height


    def small_font(self, w):
        f = max(13, int(w / 60))
        font = ImageFont.truetype(self.app.get_path('Roboto-Medium.ttf'), f)
        (width, height), (offset_x, offset_y) = font.font.getsize('A')
        return font, f, height


    def grey(self):
        if self.dark_surround:
            return (150, 150, 150, 255)
        else:
            return (50, 50, 50, 255)


    def snap_no_return(self, *args):
        self.hide()
        Clock.schedule_once(self.delayed_snap, 1)


    def delayed_snap(self, dt=None):
        try:
            if self.save_format == 'animated gif':
                Component.get('Stacker').make_animated_gif()
            else:
                self.snap()
        except:
            pass


    def snap(self, return_image=False, *args):

        self.return_image = return_image

        view = Component.get('View')
        if not hasattr(view, 'last_image') or (view.last_image is None):
            return

        dso = Component.get('DSO')

        self.save_path = '{:} {:}.{:}'.format(
                os.path.join(self.app.get_path('snapshots'), purify_name(dso.Name)), 
                datetime.now().strftime('%d%b%y_%H_%M_%S'),
                self.save_format)

        if not return_image and self.save_format == 'fits':
            self.save_as_fits()

        elif self.style == 'eyepiece':
            im = self.eyepiece_view()

        elif self.style == 'landscape':
            im = self.landscape_view()

        if return_image:
            return im


    def save_as_fits(self):
        stacker = Component.get('Stacker')
        im = stacker.get_stack()
        if im is not None:
            im = im * (2**16 - 1)
            logger.debug(f'min {np.min(im)} max {np.max(im)}')
            capture_props = {
                'exposure': stacker.describe().get('total_exposure', None),
                'sub_type': stacker.get_prop('sub_type'),
                'filt': stacker.get_prop('filter')
            }
            save_image(data=im.astype(np.uint16), path=self.save_path,
                capture_props=capture_props)
            toast(f'Saved fits to {os.path.basename(self.save_path)}', duration=2.5)
        else:
            toast('no image to save')

    # construct annotations

    def DSO_details(self):
        dso = Component.get('DSO')
        obj_details = []
        if len(dso.OT) > 0 and (dso.OT in dso.otypes):
            if len(dso.Con) > 0:
                obj_details  += [f'{dso.otypes[dso.OT]} in {dso.Con}']
            else:
                obj_details  += [format(dso.otypes[dso.OT])]

        elif len(dso.Con) > 0:
            obj_details += [dso.Con]

        if dso.RA:
            obj_details += [f'RA {str(dso.RA)}']
        if dso.Dec:
            obj_details += [f'Dec {str(dso.Dec)}']
        if dso.Mag:
            obj_details += [f'mag {dso.Mag}']
        if dso.Diam:
            obj_details += [f"diam {dso.Diam}"]

        return obj_details


    def kit_details(self):
        s = Component.get('Session')
        kit = [s.telescope, s.camera]
        ps = Component.get('PlateSolver').describe()

        fov = ps['fov']
        if fov is not None:
            kit += [fov]

        return kit


    def session_details(self):
        sesh = Component.get('Session').describe()
        ps = Component.get('PlateSolver').describe()
        fwhm = ps['fwhm']
        if fwhm is not None:
            sesh += [f'fwhm: {fwhm:.1f}"']
        return sesh


    def processing_details(self):

        block = []
        stacker = Component.get('Stacker')
        d = stacker.describe()
        if d is None:
            return []

        if d['gain'] > 0:
            block += [f"gain {d['gain']:.0f}"]

        block += [f"{Component.get('StackCombiner').combine_method} stack" if d['nsubs'] > 1 else 'single sub']
        block += [f"{Component.get('Stretcher').stretch} stretch"]

        calib = ','.join([c for c in ['dark', 'flat', 'bias'] if d[c]])
        calib = calib if calib else 'no darks/flats'
        block += [calib]

        if not d['filters'].endswith('L'):
            block += [d['filters']]

        smode = Component.get('MultiSpectral').describe_mode()
        block += [f'mode {smode}']

        if 'total_exposure' in d:
            composition = 'varied expos' if d['multiple_exposures'] else f"{d['nsubs']} x {s_to_minsec(d['sub_exposure'])}"
            block += [f"{s_to_minsec(d['total_exposure'])} ({composition})"]

        return block # block[::-1]  # reverse!


    def eyepiece_view(self, dt=None):


        ox, oy = Metrics.get('origin')
        radius = Metrics.get('radius')

        # thanks to callump for the use of scrn_density
        scrn_density = KivyMetrics.density
        cx = int(Window.left + ox / scrn_density)
        cy = int(Window.top + Window.height / scrn_density - oy / scrn_density)
        r = int(radius / scrn_density - 40)
        w = 2 * r

        try:
            im = self.mss_grabber(bbox=(cx - r, cy - r, cx + r, cy + r))
        except Exception as e:
            logger.exception(f'problem with screen grab ({e})') 
            toast(f'screen grab error ({e})')
            return

        blur = 17                     # radius of blurred edge in pixels

        # new from callump
        w = im.width
        r = w / 2

        # construct alpha channel to composite eyepiece circle
        # quite likely some of these steps are redundant
        alpha = Image.new('L', (w, w), color=0)
        _alpha = np.ones((w, w))
        grid_x, grid_y = np.meshgrid(np.arange(w), np.arange(w))
        d = pow((grid_x - r) ** 2 + (grid_y - r) ** 2, .5) - (r - 30)
        db = d <= blur
        _alpha[db] = d[db] / blur
        alpha.putdata(_alpha.ravel(), scale=255)
        if self.dark_surround:
            grad_im = Image.new('RGBA', (w, w), color=(0, 0, 0))
        else:
            grad_im = Image.new('RGBA', (w, w), color=(255, 255, 255)) 
        grad_im.putalpha(alpha)

        # composite with scope image, converting if needed (to test)
        try:
            if im.mode != 'RGBA':
                im = im.convert('RGBA')
            im = Image.alpha_composite(im, grad_im)
        except Exception as e:
            logger.exception(f'problem compositing ({e})')
            toast(f'compositing error ({e})')
            return

        # annotate Draw object with required level of detail
        sf, sf_size, hh = self.small_font(w)
        lf, lf_size, hh = self.large_font(w)
        rowsep = sf_size + 6
        if self.annotation != 'plain':    
            draw = ImageDraw.Draw(im)
            w, h = im.size
            row = self.draw_aligned(draw, 0, 0, w, Component.get('DSO').Name, lf, 
                align='left', rowsep=lf_size+30)

        if self.annotation == 'full':
            self.draw_block(self.DSO_details(), draw, row, 0, w, sf, 
                align='left', rowsep=rowsep)
            self.draw_block(self.processing_details()[::-1], draw, h - 1.5*rowsep, 0, w, sf, 
                align='right', rowsep=-rowsep)
            self.draw_block(self.kit_details(), draw, h - 1.5*rowsep, 0, w, sf, 
                align='left', rowsep=-rowsep) # neg builds from base
            self.draw_block(self.session_details(), draw, 5, 0, w, sf, 
                align='right', rowsep=rowsep)

        # for high density screens, reduce size a little
        im = im.resize((int(im.width/self.imreduction) , int(im.height/self.imreduction)))

        if self.return_image:
            return im

        self.save(im)


    @logger.catch
    def landscape_view(self):

        # to do: better scaling of fonts based on image size

        # v0.5: ensure tmp is in snapshots dir now that user can start from anywhere
        nm = os.path.join(self.app.get_path('snapshots'), '_tmp.png')
        im = Component.get('View').last_image.copy()
        try:
            imsave(nm, im[::-1])
            orig_im = Image.open(nm)
        except Exception as e:
            logger.exception(f'load/save problem ({e})')
            toast(f'load/save error ({e})')
            return

        w, h = orig_im.size
 
        # NB this is indep of screen size, unlike eyepiece view 
        sf, sf_size, small_text_height = self.small_font(w)
        lf, lf_size, large_text_height = self.large_font(w)

        rowgap = small_text_height / 2
        obj_details = self.DSO_details()
        proc_details = self.processing_details()
        sesh_details = self.session_details() + self.kit_details()
        max_rows = max([len(obj_details), len(proc_details), len(sesh_details)]) + 1
        total_height = h
        if self.annotation != 'plain':
            total_height = int(total_height + 2 * rowgap + large_text_height)
        if self.annotation == 'full':
            total_height = int(total_height + max_rows * small_text_height + (max_rows) * rowgap + 5)

        fw = int(self.framewidth)
        im = Image.new('RGBA', (w + 2 * fw, total_height + 2 * fw), 
            color=(0, 0, 0) if self.dark_surround else (255, 255, 255))

        im.paste(orig_im, (fw, fw))
        draw = ImageDraw.Draw(im)

        w += 2 * fw
        h += 2 * fw

        # move row base to base of text
        w += large_text_height

        if self.annotation != 'plain':
            row = self.draw_aligned(draw, h , 0, w, Component.get('DSO').Name, lf, 
                align='center', rowsep=large_text_height + 2 * rowgap)

        if self.annotation == 'full':
            row += rowgap
            draw.line((5, row, w - 5, row), fill=self.grey(), width=1)
            row += (rowgap + small_text_height - 10)
            cols = [int(x) for x in orig_im.size[0] * np.linspace(0, 1, 4)]
            self.draw_block(sesh_details, draw, row,
                cols[0], cols[1], sf, align='left', rowsep=small_text_height + rowgap)
            self.draw_block(obj_details, draw, row, 
                cols[1], cols[2], sf, align='center', rowsep=small_text_height + rowgap)
            self.draw_block(proc_details, draw, row, 
                cols[2], cols[3], sf, align='right', rowsep=small_text_height + rowgap)

        # delete temporary file and save
        os.remove(nm)

        if self.return_image:
            return im
        
        self.save(im)


    def draw_aligned(self, draw, row, l, r, text, font=None, margin=15, rowsep=20, align='center'):
        # helper to draw pixel text aligned; return new row position
        # alignment is within range of cols, a pair (l, r)
        textwidth, textheight = draw.textsize(text, font)
        if align == 'left':
            col = l + margin
        elif align == 'right':
            col = r - textwidth - margin
        elif align == 'center':
            col = (l + r - textwidth) / 2
        draw.text((col, row), text, font=font, fill=self.grey())
        return row + rowsep


    def draw_block(self, block, draw, row, l, r, font, align='center', rowsep=20):
        for v in block:
            if v:
                row = self.draw_aligned(draw, row, l, r, v, font, align=align, rowsep=rowsep)


    def save(self, im):
        # save, handling any format conversions
        try:
            if self.save_format == 'png':
                im.save(self.save_path)
            elif self.save_format == 'jpg':
                bg = Image.new('RGB', im.size, (255, 255, 255))
                bg.paste(im, im)
                bg.save(self.save_path)
            toast(f'saved to {self.save_path}', duration=2)
        except Exception as e:
            logger.exception(f'problem with _save ({e})')
            toast(f'error saving ({e})', duration=2)



''' panel for central region of eyepiece; provides contents and header
	need to make it scrollable!
'''

from kivy.core.window import Window
from kivy.uix.boxlayout import BoxLayout
from kivy.lang import Builder
from kivy.properties import (
	NumericProperty,
	ObjectProperty,
	StringProperty,
)
from jocular.metrics import Metrics


Builder.load_string(
	'''

<Panel>:
	header: _header
	contents: _contents
	canvas:
		Color:
			rgba: .2, .2, .2, self.panel_opacity
		Ellipse:
			pos: self.x + dp(47) + (self.width - self.height) / 2, dp(47)
			size: self.height - dp(94), self.height - dp(94)
	orientation: 'vertical'
	pos_hint: {'center_x': 10, 'center_y': .5}
	size_hint: None, None
	# padding at top to avoid ring
	Label:
		size_hint: 1, None
		height: dp(64)
	BoxLayout:
		size_hint: 1, None
		height: dp(35)
		orientation: 'horizontal'
		Label:
			size_hint: 1, 1
		BoxLayout:
			id: _header
			size_hint: None, 1
			width: dp(200)
		Label:
			size_hint: 1, 1
	BoxLayout:
		orientation: 'horizontal'
		size_hint: 1, 1
		Label:
			size_hint: 1, 1
		BoxLayout:
			canvas.before:
				Color:
					rgb: 1, 0, 0
					a: 0
				Rectangle:
					pos: self.pos
					size: self.size
			id: _contents
			orientation: 'vertical'
			size_hint: None, 1
			width: dp(480)
		Label:
			size_hint: 1, 1

	# padding at base to avoid ring
	Label:
		size_hint: 1, None
		height: dp(89)

''')


class Panel(BoxLayout):
	''' Represents the centre panel used for config/choosers etc
	'''

	# ensure only one panel is showing at a time
	current_panel = None

	contents = ObjectProperty(None)
	header = ObjectProperty(None)
	title = StringProperty('')
	panel_opacity = NumericProperty(.4)


	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		Panel.current_panel = None


	def show(self, *args):

		if Panel.current_panel is not None:
			Panel.current_panel.hide()

		Panel.current_panel = self

		# bind mouse for this panel only (not all panels!)
		Window.bind(mouse_pos=self.on_mouse_pos)

		self.on_show()
		self.pos_hint = {'center_x': .5, 'center_y': .5}  
		width, height = Window.size
		self.width = height
		self.height = height


	def hide(self, *args):

		if Panel.current_panel is None:
			return

		# important to unbind mouse pos
		Window.unbind(mouse_pos=self.on_mouse_pos)
		Panel.current_panel = None

		self.on_hide()
		self.pos_hint = {'center_x': 10, 'center_y': .5}


	def on_mouse_pos(self, *args):
		''' if mouse is outside ring, hide panel
		'''

		if Panel.current_panel is None:
			return

		x, y = args[1]
		xc, yc = Metrics.get('origin')
		r = (((xc - x) ** 2 + (yc - y) ** 2) ** .5) / Metrics.get('outside_ring')
		if r > 1:
			self.on_leave()


	def on_show(self):
		pass


	def on_hide(self):
		pass


	def on_leave(self, *args):
		self.hide()


	def toggle(self, *args):
		if self.pos_hint['center_x'] > 1:
			self.show()
		else:
			self.hide()
